#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import random
from dbus.types import UInt32
from dbus.types import Dictionary
from typing import Any

from xpra.exit_codes import ExitCode
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.common import NotificationID, ConnectionMessage
from xpra.dbus.helper import dbus_to_native
from xpra.codecs.gstreamer.capture import Capture, capture_and_encode
from xpra.gstreamer.common import get_element_str
from xpra.codecs.image import ImageWrapper
from xpra.server.shadow.root_window_model import RootWindowModel
from xpra.server.shadow.gtk_shadow_server_base import GTKShadowServerBase
from xpra.platform.posix.fd_portal import (
    SCREENCAST_IFACE, PORTAL_SESSION_INTERFACE,
    dbus_sender_name,
    get_portal_interface, get_session_interface,
    screenscast_dbus_call, remotedesktop_dbus_call,
    AvailableSourceTypes, AvailableDeviceTypes,
)
from xpra.log import Logger

log = Logger("shadow")

session_counter: int = random.randint(0, 2 ** 24)

VIDEO_MODE = envbool("XPRA_PIPEWIRE_VIDEO_MODE", True)
VIDEO_MODE_ENCODINGS = os.environ.get("XPRA_PIPEWIRE_VIDEO_ENCODINGS", "h264,vp8,vp9,av1").split(",")


class PipewireWindowModel(RootWindowModel):
    __slots__ = ("pipewire_id", "pipewire_props")

    def __init__(self, root_window, capture, title: str, geometry, node_id: int, props: typedict):
        super().__init__(root_window=root_window, capture=capture, title=title, geometry=geometry)
        self.pipewire_id = node_id
        self.pipewire_props = props


class PortalShadow(GTKShadowServerBase):
    def __init__(self, attrs: dict[str, str]):
        # we're not using X11, so no need for this check:
        os.environ["XPRA_UI_THREAD_CHECK"] = "0"
        os.environ["XPRA_NOX11"] = "1"
        GTKShadowServerBase.__init__(self, attrs)
        self.session = None
        self.session_type = "portal desktop"
        self.session_path: str = ""
        self.session_handle: str = ""
        self.authenticating_client = None
        self.capture: Capture | None = None
        self.portal_interface = get_portal_interface()
        self.input_devices = 0
        log(f"PortalShadow({attrs}) portal_interface={self.portal_interface}")

    def get_server_mode(self) -> str:
        return "portal shadow"

    def notify_new_user(self, ss) -> None:
        log("notify_new_user() start capture")
        super().notify_new_user(ss)
        if not self._window_to_id:
            self.authenticating_client = ss
            self.create_session()

    def last_client_exited(self) -> None:
        super().last_client_exited()
        self.stop_capture()
        self.stop_session()

    def client_auth_error(self, message: str) -> None:
        self.disconnect_authenticating_client(ConnectionMessage.AUTHENTICATION_FAILED, message)

    def disconnect_authenticating_client(self, reason: ConnectionMessage, message: str) -> None:
        ac = self.authenticating_client
        if ac:
            self.authenticating_client = None
            self.disconnect_protocol(ac.protocol, reason, message)
            self.cleanup_source(ac)

    def makeRootWindowModels(self) -> list:
        log("makeRootWindowModels()")
        return []

    def makeDynamicWindowModels(self) -> list:
        log("makeDynamicWindowModels()")
        return []

    def set_keymap(self, server_source, force=False) -> None:
        raise NotImplementedError()

    def start_refresh(self, wid: int) -> None:
        self.start_capture()

    def start_capture(self) -> None:
        pass

    def setup_capture(self) -> None:
        pass

    def stop_capture(self) -> None:
        c = self.capture
        if c:
            self.capture = None
            c.clean()

    def cleanup(self) -> None:
        GTKShadowServerBase.cleanup(self)
        self.portal_interface = None

    def stop_session(self) -> None:
        s = self.session
        if not s:
            return
        self.session = None
        # https://gitlab.gnome.org/-/snippets/1122
        log(f"trying to close the session {s}")
        try:
            s.Close(dbus_interface=PORTAL_SESSION_INTERFACE)
        except Exception as e:
            log(f"ignoring error closing session {s}: {e}")

    def create_session(self) -> None:
        global session_counter
        session_counter += 1
        token = f"u{session_counter}"
        self.session_path = f"/org/freedesktop/portal/desktop/session/{dbus_sender_name}/{token}"
        options: dict[str, Any] = {
            "session_handle_token": token,
        }
        log(f"create_session() session_counter={session_counter}")
        remotedesktop_dbus_call(
            self.portal_interface.CreateSession,
            self.on_create_session_response,
            options=options)

    def on_create_session_response(self, response, results) -> None:
        r = int(response)
        res = typedict(dbus_to_native(results))
        if r:
            log("on_create_session_response%s", (response, results))
            log.error(f"Error {r} creating the session")
            log.error(" session access may have been denied")
            self.client_auth_error("session not created")
            return
        self.session_handle = res.strget("session_handle")
        log("on_create_session_response%s session_handle=%s", (r, res), self.session_handle)
        if not self.session_handle:
            log.error("Error: missing session handle creating the session")
            self.client_auth_error("no session handle")
            self.quit(ExitCode.UNSUPPORTED)
            return
        self.session = get_session_interface(self.session_path)
        self.on_session_created()

    def on_session_created(self) -> None:
        self.select_devices()

    def select_devices(self) -> None:
        log("select_devices()")
        options = {
            "types": UInt32(AvailableDeviceTypes.KEYBOARD + AvailableDeviceTypes.POINTER),
        }
        remotedesktop_dbus_call(
            self.portal_interface.SelectDevices,
            self.on_select_devices_response,
            self.session_handle,
            options=options)

    def on_select_devices_response(self, response, results) -> None:
        r = int(response)
        res = dbus_to_native(results)
        if r:
            log("on_select_devices_response%s", (response, results))
            log.error(f"Error {r} selecting screencast devices")
            self.client_auth_error("failed to select devices")
            return
        log(f"on_select_devices_response devices selected, results={res}")
        self.select_sources()

    def select_sources(self) -> None:
        options = {
            "multiple": self.multi_window,
            "types": UInt32(AvailableSourceTypes.WINDOW | AvailableSourceTypes.MONITOR),
        }
        log(f"calling SelectSources with options={options}")
        screenscast_dbus_call(
            self.portal_interface.SelectSources,
            self.on_select_sources_response,
            self.session_handle,
            options=options)

    def on_select_sources_response(self, response, results) -> None:
        r = int(response)
        res = typedict(dbus_to_native(results))
        if r:
            log("on_select_sources_response%s", (response, results))
            log.error(f"Error {r} selecting screencast sources")
            self.client_auth_error("failed to select screencast sources")
            return
        log(f"on_select_sources_response sources selected, results={res}")
        self.portal_start()

    def portal_start(self) -> None:
        log("portal_start()")
        remotedesktop_dbus_call(
            self.portal_interface.Start,
            self.on_start_response,
            self.session_handle,
            "")

    def on_start_response(self, response, results) -> None:
        r = int(response)
        res = typedict(dbus_to_native(results))
        log(f"start response: {res}")
        if r:
            log("on_start_response%s", (response, results))
            log.error(f"Error {r} starting the screen capture")
            self.client_auth_error("cannot start screen capture")
            return
        streams = res.tupleget("streams")
        if not streams:
            log.error("Error: failed to start capture:")
            log.error(" missing streams")
            self.client_auth_error("no streams")
            return
        log(f"on_start_response starting pipewire capture for {streams}")
        for node_id, props in streams:
            self.start_pipewire_capture(int(node_id), typedict(props))
        self.input_devices = res.intget("devices")
        if not self.input_devices and not self.readonly:
            # ss.notify("", nid, "Xpra", 0, "", title, body, [], {}, 10*1000, icon)
            log.warn("Warning: no input devices,")
            log.warn(" keyboard and pointer events cannot be forwarded")

    def create_capture_pipeline(self, fd: int, node_id: int, w: int, h: int) -> Capture:
        capture_element = get_element_str("pipewiresrc", {
            "fd": fd,
            "path": str(node_id),
            "do-timestamp": True,
        })
        c = self.authenticating_client
        if VIDEO_MODE:
            encoding = getattr(c, "encoding", "")
            encs = getattr(c, "core_encodings", ())
            full_csc_modes = getattr(c, "full_csc_modes", {})
            log(f"create_capture_pipeline() core_encodings={encs}, full_csc_modes={full_csc_modes}")
            pipeline = capture_and_encode(capture_element, encoding, full_csc_modes, w, h)
            if pipeline:
                return pipeline
            log.warn("Warning: falling back to slow RGB capture")
        return Capture(capture_element, pixel_format="BGRX", width=w, height=h)

    def start_pipewire_capture(self, node_id: int, props: typedict) -> None:
        log(f"start_pipewire_capture({node_id}, {props})")
        if not isinstance(node_id, int):
            raise ValueError(f"node-id is a {type(node_id)}, must be an int")
        x, y = props.inttupleget("position", (0, 0))
        w, h = props.inttupleget("size", (0, 0))
        if w <= 0 or h <= 0:
            raise ValueError(f"invalid dimensions: {w}x{h}")
        empty_dict = Dictionary(signature="sv")
        fd_object = self.portal_interface.OpenPipeWireRemote(
            self.session_handle,
            empty_dict,
            dbus_interface=SCREENCAST_IFACE)
        fd = fd_object.take()
        self.capture = self.create_capture_pipeline(fd, node_id, w, h)
        self.capture.node_id = node_id
        self.capture.connect("state-changed", self.capture_state_changed)
        self.capture.connect("error", self.capture_error)
        self.capture.connect("new-image", self.capture_new_image)
        self.capture.start()
        source_type = props.intget("source_type")
        title = f"{AvailableSourceTypes(source_type)} {node_id}"
        geometry = (x, y, w, h)
        model = PipewireWindowModel(self.root, self.capture, title, geometry, node_id, props)
        # must be called from the main thread:
        log(f"new model: {model}")
        self.do_add_new_window_common(node_id, model)
        self._send_new_window_packet(model)

    def capture_new_image(self, capture, coding: str, data, client_info: dict) -> None:
        wid = capture.node_id
        model = self._id_to_window.get(wid)
        log(f"capture_new_image({capture}, {coding}, {type(data)}, {client_info}) model({wid})={model}")
        if not model:
            log.error(f"Error: cannot find window model for node {wid}")
            return
        if isinstance(data, ImageWrapper):
            self.refresh_window(model)
            return
        if not isinstance(data, bytes):
            log.warn(f"Warning: unexpected image datatype: {type(data)}")
            return
        # this is a frame from a compressed stream,
        # send it to all the window sources for this window:
        for ss in tuple(self._server_sources.values()):
            if not hasattr(ss, "get_window_source"):
                # client is not showing any windows
                continue
            ws = ss.get_window_source(wid)
            if not ws:
                # client not showing this window
                continue
            ws.direct_queue_draw(coding, data, client_info)

    def capture_error(self, capture, message) -> None:
        wid = capture.node_id
        log(f"capture_error({capture}, {message}) wid={wid}")
        log.error("Error capturing screen:")
        log.estr(message)
        model = self._id_to_window.get(wid)
        if model:
            self._remove_window(model)
        for ss in tuple(self._server_sources.values()):
            ss.may_notify(NotificationID.FAILURE, "Session Capture Failed", str(message))

    def capture_state_changed(self, capture, state) -> None:
        wid = capture.node_id
        log(f"screencast capture state changed for model {wid}: {state!r}")



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import

Gdk = gi_import("Gdk")

cursor_names = {}
cursor_types = {}


def _init_map() -> None:
    for x in dir(Gdk.CursorType):
        if not x.isupper():
            # probably a method
            continue
        try:
            v = int(getattr(Gdk.CursorType, x))
            cursor_names[v] = x
            cursor_types[x] = v
        except (TypeError, ValueError):
            pass


_init_map()


def get_default_cursor() -> Gdk.Cursor:
    display = Gdk.Display.get_default()
    return Gdk.Cursor.new_from_name(display, "default")


def main():
    # pylint: disable=import-outside-toplevel
    from xpra.util.str_fn import csv
    from xpra.platform import program_context
    with program_context("Cursors", "Cursors"):
        print(csv(cursor_types))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>

import sys
import time
import random

def print_text(text, words_per_minute=80):
	if not text:
		return
	n_chars = len(text)
	n_words = len(text.split())
	chars_per_word = float(n_chars)/n_words
	chars_per_second = (chars_per_word*words_per_minute)/60.0
	#print("characters per word: %s" % chars_per_word)
	#print("words per minute: %s" % words_per_minute)
	#print("chars per second: %s" % chars_per_second)
	for char in text:
		sys.stdout.write(char)
		sys.stdout.flush()
		pause = 1.0/chars_per_second*(2.0*random.random())
		time.sleep(pause)

def main():
	text = "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
	while True:
		print_text(text)

if __name__ == "__main__":
	main()



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# Memory buffers functions:
#
# 1) Buffer code found here and also similar to the Cython docs:
#    http://stackoverflow.com/a/28166272/428751
#    Allows to return a malloced python buffer,
#    which will be freed when the python object is garbage collected
#    (also uses memalign to allocate the buffer)
# 2) object to buffer conversion utility functions,

#cython: wraparound=False

from cpython.buffer cimport PyBuffer_FillInfo    # pylint: disable=syntax-error
from libc.stdlib cimport free
from libc.string cimport memset, memcpy
from libc.stdint cimport uintptr_t


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "memalign.h":
    void *xmemalign(size_t size) nogil
    int MEMALIGN_ALIGNMENT


cdef void free_buf(const void *p, size_t l, void *arg) noexcept nogil:
    free(<void *>p)


cdef MemBuf getbuf(size_t l, int readonly=1):
    cdef const void *p = xmemalign(l)
    if p == NULL:
        raise RuntimeError(f"failed to allocate {l} bytes of memory")
    return MemBuf_init(p, l, &free_buf, NULL, readonly)


cdef MemBuf padbuf(size_t l, size_t padding, int readonly=1):
    cdef const void *p = xmemalign(l+padding)
    if p == NULL:
        raise RuntimeError(f"failed to allocate {l} bytes of memory")
    return MemBuf_init(p, l, &free_buf, NULL, readonly)


cdef MemBuf makebuf(void *p, size_t l, int readonly=1):
    if p == NULL:
        raise ValueError(f"invalid NULL buffer pointer")
    return MemBuf_init(p, l, &free_buf, NULL, readonly)


cdef void *memalign(size_t size) noexcept nogil:
    return xmemalign(size)


def get_membuf(size_t l, int readonly=1) -> MemBuf:
    return getbuf(l, readonly)


cdef class MemBuf:
    def __len__(self):
        return self.l

    def __repr__(self):
        return "MemBuf(%#x)" % (<uintptr_t> self.p)

    cdef const void *get_mem(self):
        return self.p

    def get_mem_ptr(self):
        return <uintptr_t> self.p

    def __getbuffer__(self, Py_buffer *view, int flags):
        PyBuffer_FillInfo(view, self, <void *>self.p, self.l, self.readonly, flags)

    def __releasebuffer__(self, Py_buffer *view):
        pass

    def __dealloc__(self):
        cdef dealloc_callback *cb = self.dealloc_cb_p
        if cb != NULL:
            self.dealloc_cb_p = NULL
            cb(self.p, self.l, self.dealloc_cb_arg)


# Call this instead of constructing a MemBuf directly.  The __cinit__
# and __init__ methods can only take Python objects, so the real
# constructor is here.  See:
# https://mail.python.org/pipermail/cython-devel/2012-June/002734.html
cdef MemBuf MemBuf_init(const void *p, size_t l,
                        dealloc_callback *dealloc_cb_p,
                        void *dealloc_cb_arg,
                        int readonly=1):
    cdef MemBuf ret = MemBuf()
    ret.readonly = readonly
    ret.p = p
    ret.l = l
    ret.dealloc_cb_p = dealloc_cb_p
    ret.dealloc_cb_arg = dealloc_cb_arg
    return ret


cdef class BufferContext:
    cdef Py_buffer py_buf
    cdef object obj

    def __init__(self, obj):
        if not obj:
            raise ValueError(f"invalid buffer object {obj!r} evaluates to False ({type(obj)}")
        self.obj = obj
        memset(&self.py_buf, 0, sizeof(Py_buffer))

    def __enter__(self):
        if self.py_buf.buf != NULL:
            raise RuntimeError("invalid state: buffer has already been obtained")
        if PyObject_GetBuffer(self.obj, &self.py_buf, PyBUF_ANY_CONTIGUOUS):
            raise RuntimeError(f"failed to access buffer of {type(self.obj)}")
        return self

    def __exit__(self, *_args):
        if self.py_buf.buf == NULL:
            raise RuntimeError("invalid state: no buffer")
        PyBuffer_Release(&self.py_buf)

    def is_readonly(self) -> bool:
        if self.py_buf.buf == NULL:
            raise RuntimeError("invalid state: no buffer")
        return bool(self.py_buf.readonly)

    def __int__(self):
        if self.py_buf.buf == NULL:
            raise RuntimeError("invalid state: no buffer")
        return int(<uintptr_t> self.py_buf.buf)

    def __len__(self):
        return self.py_buf.len

    def __repr__(self):
        return "BufferContext(%s)" % self.obj


cdef class MemBufContext:
    cdef MemBuf membuf

    def __init__(self, membuf):
        if not isinstance(membuf, MemBuf):
            raise ValueError(f"{membuf!r} is not a MemBuf instance: {type(membuf)}")
        self.membuf = membuf

    def is_readonly(self) -> bool:
        return self.membuf.readonly

    def __enter__(self):
        return self

    def __exit__(self, *_args):
        self.membuf = None

    def __int__(self):
        return self.membuf.get_mem_ptr()

    def __len__(self):
        return len(self.membuf)

    def __repr__(self):
        return "MemBufContext(%s)" % self.membuf


cdef object buffer_context(object obj):
    if obj is None:
        raise ValueError(f"no buffer")
    if len(obj)==0:
        raise ValueError(f"empty {type(obj)} buffer")
    if isinstance(obj, MemBuf):
        return MemBufContext(obj)
    return BufferContext(obj)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time
from ctypes import (
    get_last_error, windll,  # @UnresolvedImport
    Structure, byref, c_ubyte, c_char,
)
from io import BytesIO
from typing import Any
from PIL import Image

from xpra.log import Logger, consume_verbose_argv
from xpra.common import roundup
from xpra.util.env import envbool
from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.gui import get_virtualscreenmetrics
from xpra.codecs.image import ImageWrapper

# user32:
from xpra.platform.win32.common import (
    GetDesktopWindow, GetWindowDC, ReleaseDC, DeleteDC,
    CreateCompatibleDC, CreateCompatibleBitmap,
    GetBitmapBits, SelectObject, DeleteObject,
    BitBlt, GetDeviceCaps,
    GetSystemPaletteEntries,
)

log = Logger("shadow", "win32")

NULLREGION = 1  # The region is empty.
SIMPLEREGION = 2  # The region is a single rectangle.
COMPLEXREGION = 3  # The region is more than a single rectangle.
REGION_CONSTS: dict[int, str] = {
    NULLREGION: "the region is empty",
    SIMPLEREGION: "the region is a single rectangle",
    COMPLEXREGION: "the region is more than a single rectangle",
}
DISABLE_DWM_COMPOSITION = envbool("XPRA_DISABLE_DWM_COMPOSITION", False)


class PALETTEENTRY(Structure):
    _fields_ = [
        ('peRed', c_ubyte),
        ('peGreen', c_ubyte),
        ('peBlue', c_ubyte),
        ('peFlags', c_ubyte),
    ]


DWM_EC_DISABLECOMPOSITION = 0
DWM_EC_ENABLECOMPOSITION = 1


def set_dwm_composition(value=DWM_EC_DISABLECOMPOSITION) -> bool:
    try:
        windll.dwmapi.DwmEnableComposition(value)
        log("DwmEnableComposition(%s) succeeded", value)
        return True
    except Exception as e:
        log.error("Error: cannot change dwm composition:")
        log.estr(e)
        return False


def get_desktop_bit_depth() -> int:
    desktop_wnd = GetDesktopWindow()
    dc = GetWindowDC(desktop_wnd)
    assert dc, "failed to get a drawing context from the desktop window %s" % desktop_wnd
    bit_depth = GetDeviceCaps(dc, win32con.BITSPIXEL)
    log("get_desktop_bit_depth()=%i", bit_depth)
    ReleaseDC(desktop_wnd, dc)
    return bit_depth


def get_palette(dc) -> list:
    count = GetSystemPaletteEntries(dc, 0, 0, None)
    log("palette size: %s", count)
    palette = []
    if count > 0:
        buf = (PALETTEENTRY * count)()
        r = GetSystemPaletteEntries(dc, 0, count, byref(buf))
        for i in range(min(count, r)):
            p = buf[i]
            # we expect 16-bit values, so bit-shift them:
            palette.append((p.peRed << 8, p.peGreen << 8, p.peBlue << 8))
    return palette


RGB_FORMATS = {
    32: "BGRX",
    30: "r210",
    24: "BGR",
    16: "BGR565",
    8: "RLE8",
}


class GDICapture:

    def __init__(self):
        self.metrics = None
        self.wnd, self.dc, self.memdc = None, None, None
        self.bit_depth = 32
        self.bitblt_err_time = 0.0
        self.disabled_dwm_composition = DISABLE_DWM_COMPOSITION and set_dwm_composition(DWM_EC_DISABLECOMPOSITION)

    def __repr__(self):
        return "GDICapture(%i-bits)" % self.bit_depth

    def get_info(self) -> dict[str, Any]:
        return {
            "type": "gdi",
            "depth": self.bit_depth,
        }

    def get_type(self) -> str:
        return "GDI"

    def refresh(self) -> bool:
        return True

    def clean(self) -> None:
        if self.disabled_dwm_composition:
            set_dwm_composition(DWM_EC_ENABLECOMPOSITION)
        self.clean_dc()

    def clean_dc(self) -> None:
        dc = self.dc
        wnd = self.wnd
        if dc and wnd:
            self.dc = None
            self.wnd = None
            ReleaseDC(wnd, dc)
        memdc = self.memdc
        if memdc:
            self.memdc = None
            DeleteDC(memdc)

    def get_capture_coords(self, x: int, y: int, width: int, height: int) -> tuple[int, int, int, int]:
        metrics = get_virtualscreenmetrics()
        if self.metrics is None or self.metrics != metrics:
            # new metrics, start from scratch:
            self.metrics = metrics
            self.clean()
        log("get_image%s metrics=%s", (x, y, width, height), metrics)
        dx, dy, dw, dh = metrics
        if width == 0:
            width = dw
        if height == 0:
            height = dh
        # clamp rectangle requested to the virtual desktop size:
        if x < dx:
            width -= x - dx
            x = dx
        if y < dy:
            height -= y - dy
            y = dy
        if width > dw:
            width = dw
        if height > dh:
            height = dh
        return x, y, width, height

    def get_image(self, x: int = 0, y: int = 0, width: int = 0, height: int = 0) -> ImageWrapper | None:
        start = time.time()
        x, y, width, height = self.get_capture_coords(x, y, width, height)
        if not self.dc:
            self.wnd = GetDesktopWindow()
            if not self.wnd:
                log.error("Error: cannot access the desktop window")
                log.error(" capturing the screen is not possible")
                return None
            self.dc = GetWindowDC(self.wnd)
            if not self.dc:
                log.error("Error: cannot get a drawing context")
                log.error(" capturing the screen is not possible")
                log.error(" desktop window=%#x", self.wnd)
                return None
            self.bit_depth = GetDeviceCaps(self.dc, win32con.BITSPIXEL)
            self.memdc = CreateCompatibleDC(self.dc)
            assert self.memdc, "failed to get a compatible drawing context from %s" % self.dc
        bitmap = CreateCompatibleBitmap(self.dc, width, height)
        if not bitmap:
            log.error("Error: failed to get a compatible bitmap")
            log.error(" from drawing context %#x with size %ix%i", self.dc, width, height)
            self.clean_dc()
            return None
        r = SelectObject(self.memdc, bitmap)
        if not r:
            log.error("Error: cannot select bitmap object")
            return None
        select_time = time.time()
        log("get_image up to SelectObject (%s) took %ims", REGION_CONSTS.get(r, r), (select_time - start) * 1000)
        try:
            if BitBlt(self.memdc, 0, 0, width, height, self.dc, x, y, win32con.SRCCOPY) == 0:
                e = get_last_error()
                # rate limit the error message:
                now = time.time()
                if now - self.bitblt_err_time > 10:
                    log.error("Error: failed to blit the screen, error %i", e)
                    self.bitblt_err_time = now
                return None
        except Exception as e:
            log("BitBlt error", exc_info=True)
            log.error("Error: cannot capture screen with BitBlt")
            log.estr(e)
            self.clean_dc()
            return None
        bitblt_time = time.time()
        log("get_image BitBlt took %ims", (bitblt_time - select_time) * 1000)
        rowstride = roundup(width * self.bit_depth // 8, 2)
        buf_size = rowstride * height
        # noinspection PyTypeChecker
        buftype = c_char * buf_size
        # noinspection PyCallingNonCallable
        buf = buftype()
        buf.value = b""
        log("GetBitmapBits(%#x, %#x, %#x)", bitmap, buf_size, byref(buf))
        r = GetBitmapBits(bitmap, buf_size, byref(buf))
        if not r:
            log.error("Error: failed to copy screen bitmap data")
            self.clean_dc()
            return None
        pixels: Any = buf
        if r != buf_size:
            log.warn("Warning: truncating pixel buffer, got %i bytes but expected %i", r, buf_size)
            pixels = buf[:r]
        log("get_image GetBitmapBits took %ims", (time.time() - bitblt_time) * 1000)
        DeleteObject(bitmap)
        assert pixels, "no pixels returned from GetBitmapBits"
        rgb_format = RGB_FORMATS.get(self.bit_depth)
        if not rgb_format:
            raise ValueError("unsupported bit depth: %s" % self.bit_depth)
        bpp = self.bit_depth // 8
        v = ImageWrapper(0, 0, width, height, pixels, rgb_format,
                         self.bit_depth, rowstride, bpp, planes=ImageWrapper.PACKED, thread_safe=True)
        if self.bit_depth == 8:
            palette = get_palette(self.dc)
            v.set_palette(palette)
        log("get_image%s=%s took %ims", (x, y, width, height), v, (time.time() - start) * 1000)
        return v

    def take_screenshot(self) -> tuple[int, int, str, int, bytes] | None:
        x, y, w, h = get_virtualscreenmetrics()
        image = self.get_image(x, y, w, h)
        if not image:
            return None
        assert image.get_width() == w and image.get_height() == h
        assert image.get_pixel_format() == "BGRX"
        img = Image.frombuffer("RGB", (w, h), image.get_pixels(), "raw", "BGRX", 0, 1)
        out = BytesIO()
        img.save(out, format="PNG")
        screenshot = (img.width, img.height, "png", img.width * 3, out.getvalue())
        out.close()
        return screenshot


def main(argv) -> int:
    from xpra.platform import program_context
    with program_context("Screen-Capture", "Screen Capture"):
        consume_verbose_argv(argv, "win32")
        capture = GDICapture()
        image = capture.take_screenshot()
        assert image, "failed to capture screenshot"
        from xpra.platform.paths import get_download_dir
        filename = os.path.join(get_download_dir(), "gdi-screenshot-%i.png" % time.time())
        with open(filename, "wb") as f:
            f.write(image[4])
        capture.clean()
    return 0


if __name__ == "__main__":
    import sys
    main(sys.argv)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from enum import IntEnum
from time import monotonic
from xpra.common import roundup
from xpra.util.str_fn import memoryview_to_bytes


def clone_plane(plane):
    if isinstance(plane, memoryview):
        return plane.tobytes()
    return plane[:]


class PlanarFormat(IntEnum):
    PACKED = 0
    PLANAR_2 = 2
    PLANAR_3 = 3
    PLANAR_4 = 4
    INVALID = -1


class ImageWrapper:
    PACKED: PlanarFormat = PlanarFormat.PACKED
    PLANAR_2: PlanarFormat = PlanarFormat.PLANAR_2
    PLANAR_3: PlanarFormat = PlanarFormat.PLANAR_3
    PLANAR_4: PlanarFormat = PlanarFormat.PLANAR_4
    PLANE_OPTIONS: tuple[PlanarFormat, PlanarFormat, PlanarFormat, PlanarFormat] = (
        PACKED, PLANAR_2, PLANAR_3, PLANAR_4,
    )

    def __init__(self, x: int, y: int, width: int, height: int, pixels, pixel_format: str, depth: int, rowstride,
                 bytesperpixel: int = 4, planes: PlanarFormat = PACKED, thread_safe: bool = True, palette=None,
                 full_range=True):
        self.x: int = x
        self.y: int = y
        self.target_x: int = x
        self.target_y: int = y
        self.width: int = width
        self.height: int = height
        self.pixels = pixels
        self.pixel_format = pixel_format
        self.depth: int = depth
        self.rowstride = rowstride
        self.bytesperpixel: int = bytesperpixel
        self.planes: PlanarFormat = planes
        self.thread_safe: bool = thread_safe
        self.freed: bool = False
        self.timestamp: int = int(monotonic() * 1000)
        self.palette = palette
        self.full_range = full_range
        if width <= 0 or height <= 0:
            raise ValueError(f"invalid geometry {x},{y},{width},{height}")

    def _cn(self):
        try:
            return type(self).__name__
        except AttributeError:  # pragma: no cover
            return type(self)

    def __repr__(self) -> str:
        return "%s(%s:%s:%s)" % (
            self._cn(), self.pixel_format, self.get_geometry(), getattr(self.planes, "name", self.planes),
        )

    def get_geometry(self) -> tuple[int, int, int, int, int]:
        return self.x, self.y, self.width, self.height, self.depth

    def get_x(self) -> int:
        return self.x

    def get_y(self) -> int:
        return self.y

    def get_target_x(self) -> int:
        return self.target_x

    def get_target_y(self) -> int:
        return self.target_y

    def set_target_x(self, target_x: int):
        self.target_x = target_x

    def set_target_y(self, target_y: int):
        self.target_y = target_y

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_rowstride(self):
        return self.rowstride

    def get_depth(self) -> int:
        return self.depth

    def get_bytesperpixel(self) -> int:
        return self.bytesperpixel

    def get_size(self) -> int:
        return self.rowstride * self.height

    def get_pixel_format(self) -> str:
        return self.pixel_format

    def get_pixels(self):
        return self.pixels

    def get_planes(self) -> PlanarFormat:
        return self.planes

    def get_palette(self):
        return self.palette

    def get_full_range(self) -> bool:
        return self.full_range

    def get_gpu_buffer(self):
        return None

    def has_pixels(self) -> bool:
        return bool(self.pixels)

    def is_thread_safe(self) -> bool:
        """ if True, free() and clone_pixel_data() can be called from any thread,
            if False, free() and clone_pixel_data() must be called from the same thread.
            Used by XImageWrapper to ensure X11 images are freed from the UI thread.
        """
        return self.thread_safe

    def get_timestamp(self) -> int:
        """ time in millis """
        return self.timestamp

    def set_timestamp(self, timestamp: int) -> None:
        self.timestamp = timestamp

    def set_planes(self, planes: PlanarFormat) -> None:
        self.planes = planes

    def set_rowstride(self, rowstride: int) -> None:
        self.rowstride = rowstride

    def set_pixel_format(self, pixel_format) -> None:
        self.pixel_format = pixel_format

    def set_palette(self, palette) -> None:
        self.palette = palette

    def set_full_range(self, full_range: bool) -> None:
        self.full_range = full_range

    def set_pixels(self, pixels) -> None:
        if self.freed:
            raise RuntimeError("image wrapper has already been freed")
        self.pixels = pixels

    def allocate_buffer(self, _buf_len, _free_existing=1) -> int:
        if self.freed:
            raise RuntimeError("image wrapper has already been freed")
        # only defined for XImage wrappers:
        return 0

    def may_restride(self) -> bool:
        newstride = roundup(self.width * self.bytesperpixel, 4)
        if self.rowstride > newstride:
            return self.restride(newstride)
        return False

    def restride(self, rowstride: int) -> bool:
        if self.freed:
            raise RuntimeError("image wrapper has already been freed")
        if self.planes > 0:
            # not supported yet for planar images
            return False
        pixels = self.pixels
        assert pixels, "no pixel data to restride"
        oldstride = self.rowstride
        pos = 0
        lines = []
        for _ in range(self.height):
            lines.append(memoryview_to_bytes(pixels[pos:pos + rowstride]))
            pos += oldstride
        if self.height > 0 and oldstride < rowstride:
            # the last few lines may need padding if the new rowstride is bigger
            # (usually just the last line)
            # we do this here to avoid slowing down the main loop above
            # as this should be a rarer case
            for h in range(self.height):
                i = -(1 + h)
                line = lines[i]
                if len(line) < rowstride:
                    lines[i] = line + b"\0" * (rowstride - len(line))
                else:
                    break
        self.rowstride = rowstride
        self.pixels = b"".join(lines)
        return True

    def freeze(self) -> bool:
        if self.freed:
            raise RuntimeError("image wrapper has already been freed")
        # some wrappers (XShm) need to be told to stop updating the pixel buffer
        return False

    def clone_pixel_data(self) -> None:
        if self.freed:
            raise RuntimeError("image wrapper has already been freed")
        pixels = self.pixels
        planes = self.planes
        if not pixels:
            raise ValueError("no pixel data to clone")
        if planes < 0:
            raise ValueError(f"invalid number of planes {planes}")
        if planes == 0:
            # no planes, simple buffer:
            self.pixels = clone_plane(pixels)
        else:
            self.pixels = [clone_plane(pixels[i]) for i in range(planes)]
        self.thread_safe = True
        if self.freed:  # pragma: no cover
            # could be a race since this can run threaded
            self.free()

    def get_sub_image(self, x: int, y: int, w: int, h: int):
        # raise NotImplementedError("no sub-images for %s" % type(self))
        if w <= 0 or h <= 0:
            raise ValueError(f"invalid sub-image size: {w}x{h}")
        if x + w > self.width:
            raise ValueError(f"invalid sub-image width: {x}+{w} greater than image width {self.width}")
        if y + h > self.height:
            raise ValueError(f"invalid sub-image height: {y}+{h} greater than image height {self.height}")
        if self.planes != ImageWrapper.PACKED:
            raise NotImplementedError("cannot sub-divide a planar image!")
        if x == 0 and y == 0 and w == self.width and h == self.height:
            # same dimensions, use the same wrapper
            return self
        # copy to local variables:
        pixels = self.pixels
        oldstride = self.rowstride
        pos = y * oldstride + x * self.bytesperpixel
        newstride = w * self.bytesperpixel
        lines = []
        for _ in range(h):
            lines.append(memoryview_to_bytes(pixels[pos:pos + newstride]))
            pos += oldstride
        image = ImageWrapper(self.x + x, self.y + y, w, h, b"".join(lines), self.pixel_format, self.depth, newstride,
                             planes=self.planes, thread_safe=True, palette=self.palette, full_range=self.full_range)
        image.set_target_x(self.target_x + x)
        image.set_target_y(self.target_y + y)
        return image

    def __del__(self) -> None:
        self.free()

    def __dealloc__(self):
        self.free()

    def free(self) -> None:
        if not hasattr(self, "freed"):
            return
        if not self.freed:
            self.freed = True
            self.planes = PlanarFormat.INVALID
            self.pixels = ()
            self.pixel_format = ""



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


from xpra.challenge.handler import AuthenticationHandler


class Handler(AuthenticationHandler):

    def __init__(self, **kwargs):
        self.password = kwargs.get("password", "")

    def __repr__(self):
        return "uri"

    def get_digest(self) -> str:
        return ""

    def handle(self, challenge: bytes, digest: str, prompt: str) -> str:  # pylint: disable=unused-argument
        return self.password



# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.log import Logger
from xpra.util.str_fn import strtobytes
from xpra.gtk.error import xsync
from xpra.x11.gtk.prop import prop_get, get_python_type
from xpra.x11.bindings.window import X11WindowBindings

window_bindings = X11WindowBindings()
log = Logger("x11", "filters")


def get_x11_window_value(prop, window):
    try:
        with xsync:
            xid = window.get_xid()
            x11type = window_bindings.GetWindowPropertyType(xid, prop)[0]
    except Exception:
        log("get_x11_window_value(%s, %s)", prop, window, exc_info=True)
        x11type = ""
    if x11type:
        ptype = get_python_type(x11type)
        # log("%s: %s (%s)", filter_object.property_name, x11type, ptype)
        assert ptype, "type '%s' is not handled!" % x11type
        v = prop_get(window.get_xid(), prop, ptype)
        log("prop_get(%s, %s, %s)=%s", window, prop, ptype, v)
        if v and isinstance(v, str):
            v = strtobytes(v).replace(b"\0", b"")
    else:
        v = None
    log("%s=%s (type=%s)", prop, v, x11type)
    return v


def get_window_value(filter_object, gdkwin):
    return get_x11_window_value(filter_object.property_name, gdkwin)


def get_window(filter_object, window):
    xid = window.get_property("xid")
    p = xid
    log(f"get_window({filter_object}, {window}) xid={xid:x}, recurse={filter_object.recurse}")
    WM_TRANSIENT_FOR = "WM_TRANSIENT_FOR"
    while filter_object.recurse and p:
        try:
            xid = prop_get(p, WM_TRANSIENT_FOR, "window", ignore_errors=True)
            log(f"prop_get({p:x}, {WM_TRANSIENT_FOR})={xid}")
            if not xid:
                return p
            p = xid
        except Exception:
            log(f"prop_get({p:x}, {WM_TRANSIENT_FOR})", exc_info=True)
            break
    return p


def init_x11_window_filters():
    from xpra.server.window import filters
    original_get_window_filter = filters.get_window_filter

    def get_x11_window_filter(object_name, property_name, operator, value):
        oname = object_name.lower()
        wf = original_get_window_filter(oname.replace("x11:", ""), property_name, operator, value)
        if oname.startswith("x11:"):
            # same filter but use X11 properties:
            import types
            wf.get_window = types.MethodType(get_window, wf)
            wf.get_window_value = types.MethodType(get_window_value, wf)
            log("patched methods: %s, %s", wf.get_window, wf.get_window_value)
        log("x11 get_window_filter%s=%s", (object_name, property_name, operator, value), wf)
        return wf

    filters.get_window_filter = get_x11_window_filter
    log("init_x11_window_filters() filters.get_window_filter=%s", filters.get_window_filter)



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.log import Logger

log = Logger("util", "exec")

DEF PROC_FILLSTATUS = 0x0020    #read status
DEF PROC_PID = 0x1000           #process id numbers ( 0   terminated )


cdef extern from "libproc2/pids.h":
    ctypedef enum pids_item:
        PIDS_ID_PPID

    # Opaque handle used by procps to store its buffers
    struct pids_info:
        pass

    # Reports counts of queried processes (both total, and per-state which we don't use)
    struct pids_counts:
        int total

    # Variant type containing a value for each query flag
    union pids_result_variant:
        int s_int

    struct pids_result:
        int item
        pids_result_variant result

    struct pids_stack:
        pids_result *head

    # Return buffer that reports queried process counts and actual results
    struct pids_fetch:
        pids_counts *counts
        pids_stack **stacks

    ctypedef enum pids_select_type:
        PIDS_SELECT_PID
        PIDS_SELECT_PID_THREADS
        PIDS_SELECT_UID
        PIDS_SELECT_UID_THREADS

    int procps_pids_new(pids_info **info, pids_item *items, int numitems)
    int procps_pids_unref(pids_info **info)
    pids_fetch *procps_pids_select(pids_info *info, unsigned*pids, int pidcount, int select_type)

def get_parent_pid(unsigned int pid) -> long:
    cdef pids_info *handle = NULL
    cdef pids_item selector = PIDS_ID_PPID
    if procps_pids_new(&handle, &selector, 1) != 0:
        return 0

    cdef pids_fetch *query = procps_pids_select(handle, &pid, 1, PIDS_SELECT_PID)
    cdef int retval = 0
    if query != NULL and query.counts.total == 1 and query.stacks[0].head.item == PIDS_ID_PPID:
        retval = query.stacks[0].head.result.s_int

    procps_pids_unref(&handle)
    return retval



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.net.websockets.common import verify_response_headers, make_websocket_accept_hash
from xpra.log import Logger

log = Logger("network")


class WebsocketHeaderTest(unittest.TestCase):

    def test_verify_response_headers(self):
        def f(v, key=b""):
            try:
                verify_response_headers(v, key)
            except Exception:
                pass
            else:
                raise Exception("bad header should have failed")
        f(None)
        f({"upgrade" : "not-websocket"})
        f({"upgrade" : "websocket",
           "sec-websocket-protocol" : "not-binary",
           })
        f({"upgrade" : "websocket",
           "sec-websocket-protocol" : "binary",
           "sec-websocket-accept" : "",
           })
        f({"upgrade" : "websocket",
           "sec-websocket-protocol" : "binary",
           "sec-websocket-accept" : "key",
           })
        key = b"somekey"
        verify_response_headers({
            "upgrade" : "websocket",
            "sec-websocket-protocol" : "binary",
            "sec-websocket-accept" : make_websocket_accept_hash(key),
        }, key)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import socket
import struct
import threading
from typing import Any, Union, TypeAlias, Final
from collections.abc import Callable, Sequence

from xpra.exit_codes import ExitCode
from xpra.net.compression import Compressed, Compressible, LargeStructure
from xpra.common import noop, SizedBuffer
from xpra.os_util import LINUX, FREEBSD, WIN32
from xpra.scripts.config import str_to_bool, InitExit
from xpra.util.str_fn import repr_ellipsized
from xpra.util.env import envint, envbool

logger = None


def get_logger():
    global logger
    if logger is None:
        from xpra.log import Logger
        logger = Logger("network")
    return logger


DEFAULT_PORT: Final[int] = 14500

DEFAULT_PORTS: dict[str, int] = {
    "ws": 80,
    "wss": 443,
    "ssl": DEFAULT_PORT,  # could also default to 443?
    "ssh": 22,
    "tcp": DEFAULT_PORT,
    "vnc": 5900,
    "quic": 20000,
}

HttpResponse: TypeAlias = tuple[int, dict, bytes]

PacketElement: TypeAlias = Union[
    tuple, list, dict, int, bool, str, bytes, memoryview,
    Compressible, Compressed, LargeStructure,
]

# packet type followed by attributes:
# in 3.11: tuple[str, *tuple[int, ...]]
# tuple[str, Unpack[tuple[int, ...]] for older versions

try:
    from typing import Unpack
    PacketType: TypeAlias = tuple[str, Unpack[tuple[PacketElement, ...]]]
except ImportError:  # pragma: no cover
    PacketType: TypeAlias = tuple[PacketElement, ...]

# client packet handler:
PacketHandlerType = Callable[[PacketType], None]
# server packet handler:
ServerPacketHandlerType = Callable[[Any, PacketType], None]

NetPacketType: TypeAlias = tuple[int, int, int, SizedBuffer]


class ConnectionClosedException(Exception):
    pass


MAX_PACKET_SIZE: int = envint("XPRA_MAX_PACKET_SIZE", 16 * 1024 * 1024)
FLUSH_HEADER: bool = envbool("XPRA_FLUSH_HEADER", True)
SSL_UPGRADE: bool = envbool("XPRA_SSL_UPGRADE", False)

AUTO_ABSTRACT_SOCKET = envbool("XPRA_AUTO_ABSTRACT_SOCKET", LINUX)
ABSTRACT_SOCKET_PREFIX: Final[str] = "xpra/"

SOCKET_TYPES: Sequence[str] = (
    "tcp", "ws", "wss", "ssl", "ssh", "rfb",
    "vsock", "hyperv", "socket",
    "named-pipe",
    "quic",
)

IP_SOCKTYPES: Sequence[str] = ("tcp", "ssl", "ws", "wss", "ssh", "quic")
TCP_SOCKTYPES: Sequence[str] = ("tcp", "ssl", "ws", "wss", "ssh")

URL_MODES: dict[str, str] = {
    "xpra": "tcp",
    "xpras": "ssl",
    "xpra+tcp": "tcp",
    "xpratcp": "tcp",
    "xpra+tls": "ssl",
    "xpratls": "ssl",
    "xpra+ssl": "ssl",
    "xprassl": "ssl",
    "xpra+ssh": "ssh",
    "xprassh": "ssh",
    "xpra+ws": "ws",
    "xpraws": "ws",
    "xpra+wss": "wss",
    "xprawss": "wss",
    "xpra+hyperv": "hyperv",
    "xprahyperv": "hyperv",
    "rfb": "vnc",
}

PACKET_TYPES: list[str] = [
    # generic:
    "hello",
    "challenge",
    "ssl-upgrade",
    "info", "info-response",
    # server state:
    "server-event", "startup-complete",
    "setting-change", "control",
    # network layer:
    "disconnect", "connection-lost", "gibberish", "invalid",
    # pings:
    "ping", "ping_echo",
    # file transfers:
    "open-url", "send-file", "send-data-request", "send-data-response", "ack-file-chunk", "send-file-chunk",
    # audio:
    "sound-data", "new-stream", "state-changed", "new-buffer", "cleanup", "add_data", "stop",
    # display:
    "show-desktop",
    # windows and trays:
    "new-window", "new-override-redirect", "new-tray",
    "raise-window", "initiate-moveresize", "window-move-resize", "window-resized", "window-metadata",
    "configure-override-redirect", "lost-window", "window-icon",
    "draw",
    "encodings",
    "eos", "cursor", "bell",
    # pointer motion and events:
    "pointer-position", "pointer",
    "button-action", "pointer-button",
    "pointer-grab", "pointer-ungrab",
    "input-devices",
    # keyboard:
    "set-keyboard-sync-enabled",
    "key-action", "key-repeat",
    "layout-changed", "keymap-changed",
    # webcam:
    "webcam-stop", "webcam-ack",
    # clipboard:
    "set-clipboard-enabled", "clipboard-token", "clipboard-request",
    "clipboard-contents", "clipboard-contents-none", "clipboard-pending-requests", "clipboard-enable-selections",
    # notifications:
    "notify_show", "notify_close",
]


def get_log_packets(exclude=False) -> Sequence[str]:
    lp = os.environ.get("XPRA_LOG_PACKETS")
    if not lp:
        return ()
    pt = []
    for x in lp.split(","):
        if x.startswith("-") == exclude:
            pt.append(x[int(exclude):])
    return tuple(pt)


def _may_log_packet(sending, packet_type, packet) -> None:
    if LOG_PACKET_TYPE:
        get_logger().info("%s %s (thread=%s)", "sending  " if sending else "receiving", packet_type,
                          threading.current_thread())
    if LOG_PACKETS or NOLOG_PACKETS:
        if packet_type in NOLOG_PACKETS:
            return
        if packet_type in LOG_PACKETS or "*" in LOG_PACKETS:
            s = str(packet)
            if len(s) > PACKET_LOG_MAX_SIZE:
                s = repr_ellipsized(s, PACKET_LOG_MAX_SIZE)
            get_logger().info(s)


LOG_PACKETS: Sequence[str] = ()
NOLOG_PACKETS: Sequence[str] = ()
LOG_PACKET_TYPE: bool = False
PACKET_LOG_MAX_SIZE: int = 500

may_log_packet: Callable = noop


def get_peercred(sock) -> tuple[int, int, int] | None:
    log = get_logger()
    if LINUX:
        SO_PEERCRED = 17
        try:
            creds = sock.getsockopt(socket.SOL_SOCKET, SO_PEERCRED, struct.calcsize(b'3i'))
            pid, uid, gid = struct.unpack(b'3i', creds)
            log("peer: %s", (pid, uid, gid))
            return pid, uid, gid
        except OSError as e:
            log("getsockopt", exc_info=True)
            log.error(f"Error getting peer credentials: {e}")
            return None
    elif FREEBSD:
        log.warn("Warning: peercred is not yet implemented for FreeBSD")
        # use getpeereid
        # then pwd to get the gid?
    return None


def is_request_allowed(proto, request="info", default=True) -> bool:
    try:
        options = proto._conn.options
        req_option = options.get(request, default)
    except AttributeError:
        return default
    r = str_to_bool(req_option, default)
    get_logger().debug(f"is_request_allowed%s={r}", (proto, request, default))
    return r


def init() -> None:
    global LOG_PACKETS, NOLOG_PACKETS, LOG_PACKET_TYPE, PACKET_LOG_MAX_SIZE
    LOG_PACKETS = get_log_packets()
    NOLOG_PACKETS = get_log_packets(True)
    LOG_PACKET_TYPE = envbool("XPRA_LOG_PACKET_TYPE", False)
    PACKET_LOG_MAX_SIZE = envint("XPRA_PACKET_LOG_MAX_SIZE", 500)

    global may_log_packet
    if LOG_PACKETS or NOLOG_PACKETS or LOG_PACKET_TYPE:
        may_log_packet = _may_log_packet
    else:
        may_log_packet = noop


init()


def get_ssh_port() -> int:
    port = envint("XPRA_SSH_PORT", 0)
    if 0 < port < 2**16:
        return port
    # on Linux, we can run "ssh -T | grep port"
    # but this usually requires root permissions to access /etc/ssh/sshd_config
    if WIN32:
        return 0
    return 22


def has_websocket_handler() -> bool:
    try:
        from xpra.net.websockets.handler import WebSocketRequestHandler
        assert WebSocketRequestHandler
        return True
    except ImportError:
        get_logger().debug("importing WebSocketRequestHandler", exc_info=True)
    return False


HTTP_UNSUPORTED = b"""HTTP/1.1 400 Bad request syntax or unsupported method

<head>
<title>Server Error</title>
</head>
<body>
<h1>Server Error</h1>
<p>Error code 400.
<p>Message: this port does not support HTTP requests.
<p>Error code explanation: 400 = Bad request syntax or unsupported method.
</body>
"""


def verify_hyperv_available() -> None:
    try:
        import socket
        s = socket.socket(socket.AF_HYPERV, socket.SOCK_STREAM, socket.HV_PROTOCOL_RAW)
    except (AttributeError, OSError) as e:
        raise InitExit(ExitCode.UNSUPPORTED,
                       f"hyperv sockets are not supported on this platform: {e}") from None
    else:
        s.close()


def open_html_url(html: str = "open", mode: str = "tcp", bind: str = "127.0.0.1") -> None:
    from xpra.log import Logger
    log = Logger("http", "network")
    log("open_html_url%s", (html, mode, bind))
    import urllib
    result = urllib.parse.urlsplit(f"//{bind}")
    host = result.hostname
    if host in ("0.0.0.0", "*"):
        host = "localhost"
    elif host == "::":
        host = "::1"
    port = result.port or DEFAULT_PORTS.get(mode)
    ssl = mode in ("wss", "ssl")
    url = "https" if ssl else "http"
    url += f"://{host}"
    if (ssl and port != 443) or (not ssl and port != 80):
        url += f":{port}"
    url += "/"
    from subprocess import Popen, SubprocessError
    from xpra.util.env import get_saved_env

    def exec_open(*cmd) -> None:
        log(f"exec_open{cmd}")
        proc = Popen(args=cmd, env=get_saved_env())
        from xpra.util.child_reaper import getChildReaper
        getChildReaper().add_process(proc, "open-html5-client", " ".join(cmd), True, True)

    def webbrowser_open() -> None:
        log.info(f"opening html5 client using URL {url!r}")
        from xpra.os_util import POSIX, OSX
        if POSIX and not OSX:
            saved_env = get_saved_env()
            if not (saved_env.get("DISPLAY") or saved_env.get("WAYLAND_DISPLAY")):
                log.warn(" no display, cannot open a browser window")
                return
            # run using a subprocess,
            # so we can specify the environment:
            # (which will run it against the correct X11 display!)
            try:
                exec_open(f"python{sys.version_info.major}", "-m", "webbrowser", "-t", url)
            except SubprocessError:
                log("failed exec_open:", exc_info=True)
            else:
                return
        import webbrowser
        webbrowser.open_new_tab(url)

    if html.lower() not in ("open", "connect"):
        # is a command?
        from xpra.util.io import which
        open_cmd = which(html)
        if open_cmd:
            log.info(f"opening html5 client using {html!r} at URL {url!r}")
            exec_open(open_cmd, url)
            return
        # fall through to webbrowser:
        log.warn(f"Warning: {html!r} is not a valid command")
    webbrowser_open()



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.auth.sys_auth_base import SysAuthenticator
from xpra.util.objects import typedict


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.salt = None

    def requires_challenge(self) -> bool:
        return False

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        return b"", ""

    def get_password(self) -> str:
        return ""

    def authenticate(self, _caps: typedict) -> bool:
        return True

    def __repr__(self):
        return "none"



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Final

CONNECTION_LOST: Final[str] = "connection-lost"
GIBBERISH: Final[str] = "gibberish"
INVALID: Final[str] = "invalid"



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: language_level=3

cdef extern from "gtk-3.0/gdk/gdk.h":
    ctypedef struct GdkDisplay:
        pass
    ctypedef struct GdkWindow:
        pass

cdef extern from "glib-2.0/glib-object.h":
    ctypedef struct cGObject "GObject":
        pass

cdef object wrap(cGObject * contents)
cdef cGObject *unwrap(box, pyclass) except? NULL



#!/usr/bin/env python3

from xpra.x11.gtk.display_source import init_gdk_display_source

init_gdk_display_source()

import gi

gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # pylint: disable=wrong-import-position @UnresolvedImport

from xpra.x11.gtk.prop import prop_set
from xpra.gtk.error import xsync


def main():
    win = Gtk.Window()
    win.set_size_request(400, 100)
    win.set_title("quality / speed hint")
    vbox = Gtk.VBox()
    win.add(vbox)

    def set_int_prop(prop_name, value):
        with xsync:
            prop_set(win.get_window(), prop_name, "u32", value)
        print("%s=%s" % (prop_name, value))

    win.quality = 100
    win.speed = 100

    def set_quality_hint():
        set_int_prop("_XPRA_QUALITY", win.quality)

    def set_speed_hint():
        set_int_prop("_XPRA_SPEED", win.speed)

    def change_quality(*_args):
        win.quality = (win.quality - 20) % 100
        set_quality_hint()

    def change_speed(*_args):
        win.speed = (win.speed - 20) % 100
        set_speed_hint()

    def add_button(label, cb):
        btn = Gtk.Button(label=label)
        vbox.add(btn)
        btn.connect('button-press-event', cb)

    add_button("change quality", change_quality)
    add_button("change speed", change_speed)
    win.connect("destroy", Gtk.main_quit)
    win.show_all()
    Gtk.main()
    return 0


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.common import SizedBuffer
#cython: wraparound=False, boundscheck=False

from libc.stdint cimport uint32_t, uintptr_t   # pylint: disable=syntax-error
from xpra.buffers.membuf cimport getbuf, MemBuf
from libc.string cimport memset


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS


def xor_str(a: SizedBuffer, b: SizedBuffer) -> memoryview:
    if len(a) != len(b):
        raise ValueError("cyxor cannot xor strings of different lengths (%s:%s vs %s:%s)" % (type(a), len(a), type(b), len(b)))
    cdef Py_buffer py_bufa
    memset(&py_bufa, 0, sizeof(Py_buffer))
    if PyObject_GetBuffer(a, &py_bufa, PyBUF_ANY_CONTIGUOUS):
        raise ValueError(f"failed to read pixel data from {type(a)}")
    cdef Py_buffer py_bufb
    memset(&py_bufb, 0, sizeof(Py_buffer))
    if PyObject_GetBuffer(b, &py_bufb, PyBUF_ANY_CONTIGUOUS):
        PyBuffer_Release(&py_bufa)
        raise ValueError(f"failed to read pixel data from {type(b)}")
    cdef Py_ssize_t alen = py_bufa.len
    cdef Py_ssize_t blen = py_bufb.len
    if alen != blen:
        PyBuffer_Release(&py_bufa)
        PyBuffer_Release(&py_bufb)
        raise RuntimeError(f"python or cython bug? buffers don't have the same length?")
    cdef MemBuf out_buf = getbuf(alen)
    cdef uintptr_t op = <uintptr_t> out_buf.get_mem()
    cdef unsigned char *acbuf = <unsigned char *> py_bufa.buf
    cdef unsigned char *bcbuf = <unsigned char *> py_bufb.buf
    cdef unsigned char *ocbuf = <unsigned char *> op
    cdef uint32_t *obuf = <uint32_t*> op
    cdef uint32_t *abuf = <uint32_t*> py_bufa.buf
    cdef uint32_t *bbuf = <uint32_t*> py_bufb.buf
    cdef unsigned int i, j, steps, char_steps
    if (alen % 4) != 0:
        #unaligned access, use byte at a time slow path:
        char_steps = alen
        for 0 <= i < char_steps:
            ocbuf[i] = acbuf[i] ^ bcbuf[i]
    else:
        #do 4 bytes at a time:
        steps = alen // 4
        if steps>0:
            for 0 <= i < steps:
                obuf[i] = abuf[i] ^ bbuf[i]
        #bytes at a time again at the end:
        char_steps = alen % 4
        if char_steps>0:
            for 0 <= i < char_steps:
                j = alen-char_steps+i
                ocbuf[j] = acbuf[j] ^ bcbuf[j]
    PyBuffer_Release(&py_bufa)
    PyBuffer_Release(&py_bufb)
    return memoryview(out_buf).toreadonly()



#!/usr/bin/env python3
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.util.env import IgnoreWarningsContext
from xpra.common import MoveResize, MOVERESIZE_DIRECTION_STRING
from xpra.gtk.window import add_close_accel
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.platform import program_context

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

width = 400
height = 400


def make_window():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_title("Window Move Resize")
    window.set_position(Gtk.WindowPosition.CENTER)
    window.connect("delete_event", Gtk.main_quit)
    icon = get_icon_pixbuf("windows.png")
    if icon:
        window.set_icon(icon)

    def get_pointer():
        with IgnoreWarningsContext():
            return window.get_window().get_screen().get_root_window().get_pointer()

    def initiate(x_root: float, y_root: float, direction: MoveResize, button: int, source_indication: int):
        from xpra.x11.gtk.display_source import init_gdk_display_source
        init_gdk_display_source()
        from xpra.x11.bindings.core import X11CoreBindings
        from xpra.x11.bindings.window import constants, X11WindowBindings
        event_mask = constants["SubstructureNotifyMask"] | constants["SubstructureRedirectMask"]
        xwin = window.get_window().get_xid()
        X11Core = X11CoreBindings()
        root_xid = X11Core.get_root_xid()
        X11Core.UngrabPointer()
        X11Window = X11WindowBindings()
        X11Window.sendClientMessage(root_xid, xwin, False, event_mask, "_NET_WM_MOVERESIZE",
                                    int(x_root), int(y_root), int(direction), button, source_indication)

    def cancel():
        initiate(0, 0, MoveResize.CANCEL, 0, 1)

    def expand(widget):
        widget.set_hexpand(True)
        widget.set_vexpand(True)
        return widget

    grid = Gtk.Grid()
    grid.set_row_homogeneous(True)
    grid.set_column_homogeneous(True)

    btn = Gtk.Button(label="initiate move")
    grid.attach(expand(btn), 2, 2, 1, 1)

    def initiate_move(*_args):
        cancel()
        pos = get_pointer()
        source_indication = 1  # normal
        button = 1
        initiate(pos.x, pos.y, MoveResize.MOVE, button, source_indication)
        GLib.timeout_add(5 * 1000, cancel)

    btn.connect('button-press-event', initiate_move)

    def btn_callback(_btn, _event, direction: MoveResize):
        cancel()
        pos = get_pointer()
        source_indication = 1  # normal
        button = 1
        initiate(pos.x, pos.y, direction, button, source_indication)
        GLib.timeout_add(5 * 1000, cancel)

    def add_button(x: int, y: int, direction: MoveResize):
        btn = Gtk.Button(label=MOVERESIZE_DIRECTION_STRING[direction])
        btn.connect('button-press-event', btn_callback, direction)
        grid.attach(expand(btn), x, y, 1, 1)

    for x, y, direction in (
            (0, 0, MoveResize.SIZE_TOPLEFT),
            (1, 0, MoveResize.SIZE_TOP),
            (2, 0, MoveResize.SIZE_TOPRIGHT),
            (0, 1, MoveResize.SIZE_LEFT),
            (1, 1, MoveResize.MOVE),
            (2, 1, MoveResize.SIZE_RIGHT),
            (0, 2, MoveResize.SIZE_BOTTOMLEFT),
            (1, 2, MoveResize.SIZE_BOTTOM),
            (2, 2, MoveResize.SIZE_BOTTOMRIGHT),
    ):
        add_button(x, y, direction)
    grid.show_all()
    window.add(grid)
    window.set_size_request(width, height)
    return window


def main():
    from xpra.gtk.signals import quit_on_signals
    with program_context("initiate-moveresize", "Initiate Move-Resize"):
        w = make_window()
        w.show_all()
        add_close_accel(w, Gtk.main_quit)
        quit_on_signals("move resize test window")
        Gtk.main()
        return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.scripts.config import str_to_bool
from xpra.gtk.configure.common import run_gui
from xpra.util.config import update_config_attribute, with_config
from xpra.gtk.widget import label
from xpra.os_util import gi_import
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("util")

FEATURES = (
    ("Audio", "Audio forwarding: speaker and microphone", "xpra.audio"),
    ("Video", "Video codecs: h264, vpx, etc", "xpra.codecs.vpx"),
    # ("Webcam", "Webcam forwarding", "xpra.codecs.v4l2"),
    ("System Tray", "System tray forwarding", "xpra.client"),
    ("File transfer", "Upload and download of files to and from the server", "xpra.net"),
    ("Printing", "Printer forwarding to the client's printer", "xpra.net"),
    ("Clipboard", "Copy & Paste to and from the server", "xpra.clipboard"),
    ("Notifications", "Notifications forwarding", "xpra.notifications"),
    ("Windows", "Windows forwarding", "xpra.client.gtk3"),
    # ("Splash", "Show the splash screen GUI", "xpra.gtk.dialogs.splash")
    # ("Readonly", "Prevent any keyboard or pointer events from being forwarded", "xpra.client.gtk3"),
)


def plabel(text, tooltip="", sensitive=False, font="sans 12") -> Gtk.Label:
    lbl = label(text, tooltip=tooltip, font=font)
    lbl.set_hexpand(False)
    lbl.set_halign(Gtk.Align.START)
    lbl.set_margin_start(5)
    lbl.set_margin_end(5)
    lbl.set_sensitive(sensitive)
    return lbl


class ConfigureGUI(BaseGUIWindow):

    def __init__(self, parent: Gtk.Window | None = None):
        self.subsystem_switch: dict[str, Gtk.Switch] = {}
        super().__init__(
            "Configure Xpra's Features",
            "features.png",
            wm_class=("xpra-configure-features-gui", "Xpra Configure Features GUI"),
            default_size=(640, 500),
            header_bar=(False, False),
            parent=parent,
        )

    def populate(self) -> None:
        self.clear_vbox()
        self.add_widget(label("Configure Xpra's Features", font="sans 20"))
        self.add_text_lines((
            "Turning off subsystems can save memory,",
            "improve security by reducing the attack surface,",
            "and also make xpra start and connect faster",
            "",
        ))

        grid = Gtk.Grid()
        grid.set_margin_start(20)
        grid.set_margin_end(20)
        grid.set_row_homogeneous(True)
        grid.set_column_homogeneous(False)
        self.add_widget(grid)

        for i, (subsystem, description, module) in enumerate(FEATURES):
            import importlib
            try:
                found = bool(importlib.import_module(module))
                tooltip = ""
            except ImportError as e:
                found = False
                tooltip = f"this feature is missing: {e}"
            sub = subsystem.lower().replace(" ", "-")
            grid.attach(plabel(subsystem, tooltip, found), 0, i, 1, 1)
            grid.attach(plabel(description, tooltip, found, font="sans 10"), 1, i, 1, 1)
            switch = Gtk.Switch()
            switch.set_sensitive(False)
            grid.attach(switch, 2, i, 1, 1)
            if found:
                switch.connect("state-set", self.toggle_subsystem, sub)
                self.subsystem_switch[sub] = switch
        self.show_all()
        with_config(self.configure_switches)

    def configure_switches(self, defaults) -> bool:
        for subsystem, switch in self.subsystem_switch.items():
            value = getattr(defaults, subsystem.replace("-", "_"), None)
            log(f"configure_switches: {subsystem}={value}")
            enabled = str_to_bool(value, False)
            switch.set_sensitive(True)
            switch.set_state(enabled)
            switch.connect("state-set", self.toggle_subsystem, subsystem)
        return False

    @staticmethod
    def toggle_subsystem(_widget, state, subsystem: str) -> None:
        update_config_attribute(subsystem, bool(state))


def main(_args) -> int:
    return run_gui(ConfigureGUI)


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv[1:]))



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import uuid
from time import monotonic
from hashlib import sha1
from base64 import b64encode
from urllib.parse import quote
from typing import Any
from enum import IntEnum
from collections.abc import Callable
from importlib import import_module

from xpra.util.str_fn import strtobytes
from xpra.log import Logger

log = Logger("websocket")

MAX_WRITE_TIME = 5
MAX_READ_TIME = 5
READ_CHUNK_SIZE = 4096

HEADERS_MODULES = os.environ.get("XPRA_WEBSOCKET_HEADERS_MODULES", "default").split(",")


class OPCODE(IntEnum):
    CONTINUE = 0
    TEXT = 1
    BINARY = 2
    CLOSE = 8
    PING = 9
    PONG = 10


OPCODE_STR: dict[int, str] = {
    OPCODE.CONTINUE: "CONTINUE",
    OPCODE.TEXT: "TEXT",
    OPCODE.BINARY: "BINARY",
    OPCODE.CLOSE: "CLOSE",
    OPCODE.PING: "PING",
    OPCODE.PONG: "PONG",
}


def make_websocket_accept_hash(key: bytes) -> bytes:
    GUID = b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
    accept = sha1(key + GUID).digest()
    return b64encode(accept)


def get_headers(host: str, port: int) -> dict[bytes, bytes]:
    headers = {}
    for mod_name in HEADERS_MODULES:
        try:
            header_module = import_module(f"xpra.net.websockets.headers.{mod_name}")
            v = header_module.get_headers(host, port)
            log(f"{mod_name}.get_headers({host}, {port})={v}")
            headers.update(v)
        except ImportError as e:
            log("import %s", mod_name, exc_info=True)
            log.error(f"Error: websocket header module {mod_name!r} not available")
            log.estr(e)
        except Exception as e:
            log("get_headers %s", mod_name, exc_info=True)
            log.error(f"Error: cannot get headers from module {mod_name!r}")
            log.estr(e)
    return headers


def client_upgrade(read: Callable, write: Callable, host: str, port: int, path="") -> None:
    key = b64encode(uuid.uuid4().bytes)
    request = get_client_upgrade_request(host, port, path, key)
    write_request(write, request)
    headers = read_server_upgrade(read)
    verify_response_headers(headers, key)
    log("client_upgrade: done")


def get_client_upgrade_request(host: str, port: int, path: str, key: bytes) -> bytes:
    url_path = quote(path)
    request = f"GET /{url_path} HTTP/1.1"
    log(f"client websocket upgrade request: {request!r}")
    lines = [request.encode("latin1")]
    headers = get_headers(host, port)
    headers[b"Sec-WebSocket-Key"] = key
    for k, v in headers.items():
        lines.append(b"%s: %s" % (k, v))
    lines.append(b"")
    lines.append(b"")
    return b"\r\n".join(lines)


def write_request(write: Callable, http_request) -> None:
    now = monotonic()
    while http_request:
        elapsed = monotonic() - now
        if elapsed >= MAX_WRITE_TIME:
            raise RuntimeError(f"http write timeout, took more {elapsed:.1f} seconds")
        w = write(http_request)
        http_request = http_request[w:]


def read_server_upgrade(read: Callable) -> dict[str, str]:
    now = monotonic()
    response = b""

    def hasheader(k) -> bool:
        return k in parse_response_header(response)

    while monotonic() - now < MAX_READ_TIME and not (
            hasheader("sec-websocket-protocol") or hasheader("www-authenticate")):
        response += read(READ_CHUNK_SIZE)
    return parse_response_header(response)


def parse_response_header(response: bytes) -> dict[str, str]:
    head = response.split(b"\r\n\r\n", 1)[0]
    lines = head.split(b"\r\n")
    headers = {}
    for line in lines:
        parts = line.decode("latin1").split(": ", 1)
        if len(parts) == 2:
            headers[parts[0].lower()] = parts[1]
    return headers


def verify_response_headers(headers: dict[str, Any], key: bytes):
    log(f"verify_response_headers({headers!r}, {key!r})")
    if not headers:
        raise ValueError("no http headers found in response")
    if headers.get("www-authenticate"):
        raise ValueError("http connection requires authentication")
    upgrade = headers.get("upgrade")
    if not upgrade:
        raise ValueError("http connection was not upgraded to websocket")
    if upgrade != "websocket":
        raise ValueError(f"invalid http upgrade: {upgrade!r}")
    protocol = headers.get("sec-websocket-protocol")
    if protocol != "binary":
        raise ValueError(f"invalid websocket protocol: {protocol!r}")
    accept_key = headers.get("sec-websocket-accept")
    if not accept_key:
        raise ValueError("websocket accept key is missing")
    expected_key = make_websocket_accept_hash(key)
    if strtobytes(accept_key) != expected_key:
        log(f"expected {expected_key!r}, received {accept_key!r}")
        raise ValueError("websocket accept key is invalid")



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.gtk.configure.common import run_gui
from xpra.util.config import update_config_attribute, with_config
from xpra.util.io import which
from xpra.util.system import is_Debian, is_Ubuntu
from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.gtk.widget import label, setfont
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("gstreamer", "util")


XDUMMY_INFO = [
    "Xdummy",
    "Xorg server with the `dummy` driver",
    "runs a full `Xorg` server",
    "supports multi-monitor virtualization and DPI emulation",
]
if is_Debian() or is_Ubuntu():
    XDUMMY_INFO.append("should not be used on Debian or Ubuntu due to severe bugs")


VFB_BACKENDS: dict[str, tuple[Sequence[str], Sequence[str]]] = {
    "auto": (
        (),
        (
            "auto",
            "Automatic runtime detection",
            "this is usually a safe option",
        ),
    ),
    "Xdummy": (
        ("Xorg", ),
        XDUMMY_INFO,
    ),
    "weston+Xwayland": (
        ("weston", "Xwayland"),
        (
            "weston + Xwayland",
            "runs Xwayland on a weston server headless backend",
            "alternative to Xdummy",
            "with more limited multi-monitor and DPI support",
        ),
    ),
    "Xvfb": (
        ("Xvfb", ),
        (
            "Xvfb",
            "traditional Xvfb virtual framebuffer",
            "It emulates a dumb framebuffer using virtual memory",
            "only requires the `Xvfb` binary installed",
            "no multi-monitor support and limited DPI emulation",
        ),
    ),
    #    "Xephyr": (
    #        ("Xephyr", ),
    #        (
    #            "Xephyr",
    #            "Launches a nested display from the current X11 server",
    #            "requires an existing X11 session",
    #            "only really useful for debugging",
    #        ),
    #    ),
}


def _set_labels_text(widgets, *messages):
    for i, widget in enumerate(widgets):
        if i < len(messages):
            widget.set_text(messages[i])
        else:
            widget.set_text("")


class ConfigureGUI(BaseGUIWindow):

    def __init__(self, parent: Gtk.Window | None = None):
        self.buttons: list[Gtk.CheckButton] = []
        size = (800, 554)
        super().__init__(
            "Configure Xpra's xvfb command",
            "monitor.png",
            wm_class=("xpra-configure-vfb-gui", "Xpra Configure vfb GUI"),
            default_size=size,
            header_bar=(False, False),
            parent=parent,
        )
        self.set_resizable(False)

    def populate(self) -> None:
        with_config(self.do_populate)

    def do_populate(self, config) -> bool:
        self.clear_vbox()
        self.set_box_margin()
        self.add_widget(label("Configure Xpra's xvfb command", font="sans 20"))

        xvfb = config.xvfb
        log(f"current {xvfb=!r}")
        for backend, defs in VFB_BACKENDS.items():
            commands, details = defs
            available = True
            tooltip = ""
            matched = backend.lower().replace(" ", "") == xvfb.lower()
            for command in commands:
                if xvfb == command:
                    matched = True
                if not which(command):
                    available = False
                    tooltip = f"{command} not found"
            description = details[0]
            btn = Gtk.CheckButton(label=description)
            btn.set_tooltip_text(tooltip)
            btn.set_sensitive(available)
            btn.set_active(available and matched)
            btn.xvfb_backend = backend
            setfont(btn, font="sans 14")
            self.vbox.add(btn)
            for detail in details[1:]:
                lbl = label(detail)
                lbl.set_halign(Gtk.Align.START)
                lbl.set_margin_start(32)
                lbl.set_sensitive(available)
                self.vbox.add(lbl)
            self.buttons.append(btn)
        btn_box = Gtk.HBox(homogeneous=True, spacing=40)
        btn_box.set_vexpand(True)
        btn_box.set_valign(Gtk.Align.END)
        self.vbox.add(btn_box)
        cancel_btn = Gtk.Button.new_with_label("Cancel")
        cancel_btn.connect("clicked", self.dismiss)
        btn_box.add(cancel_btn)
        confirm_btn = Gtk.Button.new_with_label("Save")
        confirm_btn.connect("clicked", self.save_xvfb)
        confirm_btn.set_sensitive(False)
        btn_box.add(confirm_btn)

        # only enable the confirm button once an option has been chosen,
        # and ensure that there is always one option selected
        def option_toggled(toggled_btn=None, *_args) -> None:
            if toggled_btn and toggled_btn.get_active():
                for button in self.buttons:
                    if button != toggled_btn:
                        button.set_active(False)
            else:
                if not any(button.get_active() for button in self.buttons):
                    self.buttons[0].set_active(True)
            confirm_btn.set_sensitive(any(button.get_active() for button in self.buttons))

        for btn in self.buttons:
            btn.connect("toggled", option_toggled)
        option_toggled()
        self.vbox.show_all()
        return False

    def save_xvfb(self, *_args) -> None:
        active = [button for button in self.buttons if button.get_active()]
        assert len(active) == 1
        xvfb = active[0].xvfb_backend
        log.info(f"saving {xvfb=!r}")
        update_config_attribute("xvfb", xvfb)
        self.dismiss()


def main(_args) -> int:
    return run_gui(ConfigureGUI)


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import tempfile
import unittest

from contextlib import nullcontext
from xpra.util.objects import AdHocStruct, typedict
from xpra.client.mixins import mmap

from unit.test_util import silence_info, silence_error
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class MixinsTest(ClientMixinTest):

    def test_mmap(self):

        class badfile:
            def close(self):
                raise Exception("test close failure handling")

        tmp_dir = tempfile.gettempdir()
        for mmap_option, ctx in {
            "off": nullcontext(),
            "on": silence_info(mmap),
            "auto": silence_info(mmap),
            tmp_dir+"/xpra-mmap-test-file-%i" % os.getpid(): silence_info(mmap),
            tmp_dir+"/xpra-fail-mmap-test-file-%i" % os.getpid(): silence_error(mmap),
        }.items():
            opts = AdHocStruct()
            opts.mmap = mmap_option
            opts.mmap_group = False
            with ctx:
                m = self._test_mixin_class(mmap.MmapClient, opts, {
                    "mmap": {
                        "write": {
                            "enabled": True,
                        }
                    },
                })
            x = self.mixin.mmap_read_area
            # expected = mmap_option != "off" and not mmap_option.find("fail") >= 0
            # got = bool(x)
            # we can't check any more because the `enable_from_caps` method
            # now actually checks the token...
            # assert got == expected, f"expected {expected} but got {got} for {mmap_option=}"
            m.cleanup()
            # no-op:
            m.cleanup()
            if x:
                x.tempfile = badfile()
                m.cleanup()

    def make_caps(self, caps=None) -> typedict:
        d = super().make_caps(caps)
        x = self.mixin.mmap_read_area
        if x:
            index = x.token_index
            if x.filename and x.filename.find("fail") >= 0:
                index -= 10
            d["mmap"] = {
                "enabled": True,
                "token": x.token,
                "token_bytes": x.token_bytes,
                "token_index": index,
            }
        return typedict(d)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
#  pylint: disable-msg=E1101

from xpra.platform.gui import get_native_tray_classes, get_native_tray_menu_helper_class
from xpra.os_util import gi_import, WIN32, OSX
from xpra.util.objects import make_instance
from xpra.util.env import envint
from xpra.common import XPRA_APP_ID, ConnectionMessage
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("tray")

TRAY_DELAY = envint("XPRA_TRAY_DELAY", 0)


class TrayClient(StubClientMixin):
    """
    Mixin for supporting our system tray
    (not forwarding other application's trays - that's handled in WindowClient)
    """
    PREFIX = "tray"

    def __init__(self):
        # settings:
        self.tray_icon = None
        # state:
        self.tray = None
        self.menu_helper = None

    def init(self, opts) -> None:
        if not opts.tray:
            return
        self.tray_icon = opts.tray_icon
        self.get_menu_helper()
        if opts.delay_tray:
            self.connect("first-ui-received", self.setup_xpra_tray)
        else:
            if WIN32 or OSX:
                # show shortly after the main loop starts running:
                GLib.timeout_add(TRAY_DELAY, self.setup_xpra_tray)
            else:
                # wait for handshake:
                # see appindicator bug #3956
                self.after_handshake(self.setup_xpra_tray)

    def setup_xpra_tray(self, *args) -> None:
        log("setup_xpra_tray%s", args)
        tray = self.create_xpra_tray(self.tray_icon or "xpra")
        self.tray = tray
        if tray:
            tray.show()
            icon_timestamp = tray.icon_timestamp

            def reset_icon() -> None:
                if not self.tray:
                    return
                # re-set the icon after a short delay,
                # seems to help with buggy tray geometries,
                # but don't do it if we have already changed the icon
                # (ie: the dynamic window icon code may have set a new one)
                if icon_timestamp == tray.icon_timestamp:
                    tray.set_icon()

            GLib.timeout_add(1000, reset_icon)

    def cleanup(self) -> None:
        t = self.tray
        if t:
            self.tray = None
            with log.trap_error("Error during tray cleanup"):
                t.cleanup()

    def get_tray_classes(self) -> list[type]:
        # subclasses may add their toolkit specific variants, if any
        # by overriding this method
        # use the native ones first:
        return get_native_tray_classes()

    def get_menu_helper(self):
        """
        menu helper used by our tray (make_tray / setup_xpra_tray)
        and for showing the menu on windows via a shortcut,
        this method is overriden in the gtk3.client
        """
        mhc = (get_native_tray_menu_helper_class(), )
        self.menu_helper = make_instance(mhc, self)
        return self.menu_helper

    def show_menu(self, *_args) -> None:
        if self.menu_helper:
            self.menu_helper.activate()

    def create_xpra_tray(self, tray_icon_filename: str):
        tray = None

        # this is our own tray

        def xpra_tray_click(button, pressed, time=0):
            log("xpra_tray_click(%s, %s, %s)", button, pressed, time)
            if button == 1 and pressed:
                GLib.idle_add(self.menu_helper.activate, button, time)
            elif button in (2, 3) and not pressed:
                GLib.idle_add(self.menu_helper.popup, button, time)

        def xpra_tray_mouseover(*args):
            log("xpra_tray_mouseover%s", args)

        def xpra_tray_exit(*args):
            log("xpra_tray_exit%s", args)
            self.disconnect_and_quit(0, ConnectionMessage.CLIENT_EXIT)

        def xpra_tray_geometry(*args):
            if tray:
                log("xpra_tray_geometry%s geometry=%s", args, tray.get_geometry())

        menu = None
        if self.menu_helper:
            menu = self.menu_helper.build()
        tray = self.make_tray(XPRA_APP_ID, menu, self.get_tray_title(), tray_icon_filename,
                              xpra_tray_geometry, xpra_tray_click, xpra_tray_mouseover, xpra_tray_exit)
        log("setup_xpra_tray(%s)=%s (%s)", tray_icon_filename, tray, type(tray))
        if tray:
            def reset_tray_title() -> None:
                tray.set_tooltip(self.get_tray_title())

            self.after_handshake(reset_tray_title)
        return tray

    def make_tray(self, *args):
        """ tray used by our own application """
        tc = self.get_tray_classes()
        log("make_tray%s tray classes=%s", args, tc)
        return make_instance(tc, self, *args)

    def get_tray_title(self) -> str:
        t: list[str] = []
        if self.session_name or self.server_session_name:
            t.append(self.session_name or self.server_session_name)
        ce = self.get_connection_endpoint()
        if ce:
            t.append(ce)
        if not t:
            t.insert(0, "Xpra")
        v = "\n".join(str(x) for x in t if x)
        log("get_tray_title()=%r (items=%s)", v, t)
        return v



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic

from xpra.os_util import gi_import
from xpra.common import ScreenshotData
from xpra.codecs.image import ImageWrapper
from xpra.gtk.util import get_default_root_window
from xpra.gtk.pixbuf import pixbuf_save_to_memory
from xpra.log import Logger

log = Logger("shadow")

Gdk = gi_import("Gdk")


def get_rgb_rawdata(window, x: int, y: int, width: int, height: int) \
        -> tuple[int, int, int, int, bytes, str, int, int, int] | None:
    """
        Extracts pixels from the given pixmap
    """
    start = monotonic()
    pixmap_w, pixmap_h = window.get_geometry()[2:4]
    # Just in case we somehow end up with damage larger than the pixmap,
    # we don't want to start requesting random chunks of memory (this
    # could happen if a window is resized, but we don't throw away our
    # existing damage map):
    if x < 0 or y < 0:
        raise ValueError(f"invalid position {x},{y}")
    if x + width > pixmap_w:
        width = pixmap_w - x
    if y + height > pixmap_h:
        height = pixmap_h - y
    if width <= 0 or height <= 0:
        return None
    pixbuf = Gdk.pixbuf_get_from_window(window, x, y, width, height)
    log("get_rgb_rawdata(..) pixbuf.get_from_drawable took %s ms", int(1000 * (monotonic() - start)))
    raw_data = pixbuf.get_pixels()
    rowstride = pixbuf.get_rowstride()
    return x, y, width, height, raw_data, "RGB", 24, rowstride, 3


def take_png_screenshot(window) -> ScreenshotData | None:
    log("grabbing screenshot")
    w, h = window.get_geometry()[2:4]
    pixbuf = Gdk.pixbuf_get_from_window(window, 0, 0, w, h)
    if not pixbuf:
        return None
    data = pixbuf_save_to_memory(pixbuf, "png")
    rowstride = w * 3
    return w, h, "png", rowstride, data


class GTKImageCapture:
    __slots__ = "window",

    def __init__(self, window):
        self.window = window

    def __repr__(self):
        return "GTKImageCapture(%s)" % self.window

    def get_type(self):
        return "GTK"

    def clean(self):
        """ subclasses may want to perform cleanup here """

    def refresh(self) -> bool:
        return True

    def get_image(self, x: int, y: int, width: int, height: int) -> ImageWrapper | None:
        attrs = get_rgb_rawdata(self.window, x, y, width, height)
        if not attrs:
            return None
        return ImageWrapper(*attrs)

    def take_screenshot(self) -> ScreenshotData | None:
        return take_png_screenshot(self.window)


def main(filename) -> int:
    root = get_default_root_window()
    data = take_png_screenshot(root)
    assert data
    with open(filename, "wb") as f:
        f.write(data[-1])
    return 0


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print(f"usage: {sys.argv[0]} filename.png")
        v = 1
    else:
        v = main(sys.argv[1])
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Iterable

# ensure that we use gtk as display source:
from xpra.x11.gtk.display_source import init_gdk_display_source
from xpra.keyboard.mask import DEFAULT_MODIFIER_MEANINGS
from xpra.util.objects import typedict
from xpra.util.str_fn import std, csv, bytestostr, Ellipsizer
from xpra.util.env import envbool
from xpra.gtk.error import xsync, xlog
from xpra.x11.bindings.keyboard import X11KeyboardBindings
from xpra.log import Logger

init_gdk_display_source()
X11Keyboard = X11KeyboardBindings()

log = Logger("x11", "keyboard")
verboselog = Logger("x11", "keyboard", "verbose")

XKB = envbool("XPRA_XKB", True)

DEFAULT_RULES = os.environ.get("XKB_DEFAULT_RULES", "evdev")
DEFAULT_MODEL = os.environ.get("XKB_DEFAULT_MODEL", "pc105")
DEFAULT_LAYOUT = os.environ.get("XKB_DEFAULT_LAYOUT", "us")
DEFAULT_VARIANT = os.environ.get("XKB_DEFAULT_VARIANT", "")
DEFAULT_OPTIONS = os.environ.get("XKB_DEFAULT_OPTIONS", "")

DEBUG_KEYSYMS = [x for x in os.environ.get("XPRA_DEBUG_KEYSYMS", "").split(",") if len(x) > 0]

# keys we choose not to map if the free space in the keymap is too limited
# this list was generated using:
# $ DISPLAY=:1 xmodmap -pke | awk -F= '{print $2}' | xargs -n 1 echo | sort -u | grep XF | xargs
OPTIONAL_KEYS = [
    "XF86AudioForward", "XF86AudioLowerVolume", "XF86AudioMedia", "XF86AudioMicMute", "XF86AudioMute",
    "XF86AudioNext", "XF86AudioPause", "XF86AudioPlay", "XF86AudioPrev", "XF86AudioRaiseVolume",
    "XF86AudioRecord", "XF86AudioRewind", "XF86AudioStop",
    "XF86Back", "XF86Battery", "XF86Bluetooth", "XF86Calculator", "XF86ClearGrab", "XF86Close",
    "XF86Copy", "XF86Cut", "XF86Display", "XF86Documents", "XF86DOS", "XF86Eject", "XF86Explorer",
    "XF86Favorites", "XF86Finance", "XF86Forward", "XF86Game", "XF86Go", "XF86HomePage", "XF86KbdBrightnessDown",
    "XF86KbdBrightnessUp", "XF86KbdLightOnOff",
    "XF86Launch1", "XF86Launch2", "XF86Launch3", "XF86Launch4", "XF86Launch5", "XF86Launch6",
    "XF86Launch7", "XF86Launch8", "XF86Launch9", "XF86LaunchA", "XF86LaunchB",
    "XF86Mail", "XF86MailForward", "XF86MenuKB", "XF86Messenger", "XF86MonBrightnessDown",
    "XF86MonBrightnessUp", "XF86MyComputer",
    "XF86New", "XF86Next_VMode", "XF86Open", "XF86Paste", "XF86Phone", "XF86PowerOff",
    "XF86Prev_VMode", "XF86Reload", "XF86Reply", "XF86RotateWindows", "XF86Save", "XF86ScreenSaver",
    "XF86ScrollDown", "XF86ScrollUp", "XF86Search", "XF86Send", "XF86Shop", "XF86Sleep", "XF86Suspend",
    "XF86Switch_VT_1", "XF86Switch_VT_10", "XF86Switch_VT_11", "XF86Switch_VT_12", "XF86Switch_VT_2", "XF86Switch_VT_3",
    "XF86Switch_VT_4", "XF86Switch_VT_5", "XF86Switch_VT_6", "XF86Switch_VT_7", "XF86Switch_VT_8", "XF86Switch_VT_9",
    "XF86Tools", "XF86TouchpadOff", "XF86TouchpadOn", "XF86TouchpadToggle", "XF86Ungrab", "XF86WakeUp", "XF86WebCam",
    "XF86WLAN", "XF86WWW", "XF86Xfer",
]


def clean_keyboard_state() -> None:
    with xlog:
        X11Keyboard.ungrab_all_keys()
    with xlog:
        X11Keyboard.set_layout_group(0)
    with xlog:
        X11Keyboard.unpress_all_keys()

################################################################################
# keyboard layouts


def do_set_keymap(layout: str, variant: str, options, query_struct) -> None:
    """
        layout is the generic layout name (used on non posix platforms)
            defaults to `us`
        variant is the layout variant (optional)
        query_struct is the output of `setxkbmap -query` on the client
        parsed into a dictionary
        Use those to try to set up the correct keyboard map for the client
        so that all the keycodes sent will be mapped
    """
    # First we try to use data from setxkbmap -query,
    # preferably as structured data:
    query_struct = typedict(query_struct)
    rules = DEFAULT_RULES
    model = DEFAULT_MODEL
    layout = layout or DEFAULT_LAYOUT
    if query_struct:
        log("do_set_keymap using xkbmap_query struct=%s", query_struct)
        # The query_struct data will look something like this:
        #    {
        #    b"rules"       : b"evdev",
        #    b"model"       : b"pc105",
        #    b"layout"      : b"gb",
        #    b"options"     : b"grp:shift_caps_toggle",
        #    }
        # parse the data into a dict:
        rules = query_struct.strget("rules", DEFAULT_RULES)
        model = query_struct.strget("model", DEFAULT_MODEL)
        layout = query_struct.strget("layout", DEFAULT_LAYOUT)
        variant = query_struct.strget("variant", DEFAULT_VARIANT)
        options = query_struct.strget("options", DEFAULT_OPTIONS)
        if layout:
            log.info("setting keymap: %s",
                     csv(f"{k}={std(v)}" for k, v in {
                         "rules": rules,
                         "model": model,
                         "layout": layout,
                         "variant": variant,
                         "options": options,
                     }.items() if v))
            if safe_setxkbmap(rules, model, layout, variant, options):
                return
        else:
            if safe_setxkbmap(rules, model, "", "", ""):
                return
    # fallback for non X11 clients:
    log.info("setting keyboard layout to %r", std(layout))
    safe_setxkbmap(rules, model, layout, variant, options)


def safe_setxkbmap(rules: str, model: str, layout: str, variant: str, options: str):
    # (we execute the options separately in case that fails..)
    try:
        if X11Keyboard.setxkbmap(rules, model, layout, variant, options):
            return True
    except Exception:
        log("safe_setxkbmap%s", (rules, model, layout, variant, options), exc_info=True)
        log.warn("Warning: failed to set exact keymap,")
        log.warn(f" {rules=}, {model=}, {layout=}, {variant=}, {options=}")
    if options:
        # try again with no options:
        try:
            X11Keyboard.setxkbmap(rules, model, layout, variant, "")
            return True
        except Exception:
            log("setxkbmap", exc_info=True)
            log.error("Error: failed to set exact keymap")
            log.error(" even without applying any options..")
    return False

################################################################################
# keycodes


def apply_xmodmap(instructions: list[tuple]) -> list[tuple]:
    try:
        with xsync:
            unset = X11Keyboard.set_xmodmap(instructions)
    except Exception as e:
        log("apply_xmodmap(%s)", instructions, exc_info=True)
        log.error("Error configuring modifier map: %s", e)
        unset = instructions
    if unset is None:
        # None means an X11 error occurred, re-do all:
        unset = instructions
    return unset


def get_keycode_mappings() -> dict[str, list[str]]:
    if XKB and X11Keyboard.hasXkb():
        return X11Keyboard.get_xkb_keycode_mappings()
    return X11Keyboard.get_keycode_mappings()


def set_keycode_translation(xkbmap_x11_keycodes, xkbmap_keycodes) -> dict:
    """
        Translate the given keycodes into the existing keymap
    """
    # get the list of keycodes (either from x11 keycodes or gtk keycodes):
    if xkbmap_x11_keycodes:
        dump_dict(xkbmap_x11_keycodes)
        keycodes = indexed_mappings(xkbmap_x11_keycodes)
    else:
        keycodes = gtk_keycodes_to_mappings(xkbmap_keycodes)
    x11_keycodes = get_keycode_mappings()
    log("set_keycode_translation(%s, %s)", Ellipsizer(xkbmap_x11_keycodes), Ellipsizer(xkbmap_keycodes))
    log(" keycodes=%s", Ellipsizer(keycodes))
    log(" x11_keycodes=%s", Ellipsizer(x11_keycodes))
    verboselog("set_keycode_translation(%s, %s)", xkbmap_x11_keycodes, xkbmap_keycodes)
    verboselog(" keycodes=%s", keycodes)
    verboselog(" x11_keycodes=%s", x11_keycodes)
    """
    Example data:
    ```
    keycodes = {
        9: set([('', 1), ('Escape', 4), ('', 3), ('Escape', 0), ('Escape', 2)]),
        10: set([('onesuperior', 4), ('onesuperior', 8), ('exclam', 1), ('1', 6),
                 ('exclam', 3), ('1', 2), ('exclamdown', 9), ('exclamdown', 5), ('1', 0), ('exclam', 7)]),

    x11_keycodes = {
       8: ['Mode_switch', '', 'Mode_switch', '', 'Mode_switch'],
       9: ['Escape', '', 'Escape', '', 'Escape'],
       }
    ```
    create faster lookup table:
    """
    x11_keycodes_for_keysym = {}
    for keycode, keysyms in x11_keycodes.items():
        for keysym in keysyms:
            x11_keycodes_for_keysym.setdefault(keysym, set()).add(keycode)

    def find_keycode(kc: int, keysym, i: int) -> tuple | None:
        keycodes = tuple(x11_keycodes_for_keysym.get(keysym, set()))
        if keysym in DEBUG_KEYSYMS:
            log.info("set_keycode_translation: find_keycode%s x11 keycodes=%s", (kc, keysym, i), keycodes)

        def rlog(keycode, msg) -> None:
            if keysym in DEBUG_KEYSYMS:
                log.info("set_keycode_translation: find_keycode%s=%s (%s)", (kc, keysym, i), keycode, msg)
        if not keycodes:
            return None
        # no other option, use it:
        for keycode in keycodes:
            defs = x11_keycodes.get(keycode)
            if keysym in DEBUG_KEYSYMS:
                log.info("server x11 keycode %i: %s", keycode, defs)
            if not defs:
                raise RuntimeError(f"bug: keycode {keycode} not found in {x11_keycodes}")
            if len(defs) > i and defs[i] == keysym:
                rlog(keycode, "exact index match")
                return keycode, True
        # if possible, use the same one:
        if kc in keycodes:
            rlog(kc, "using same keycode as client")
            return kc, False
        keycode = keycodes[0]
        rlog(keycode, "using first match")
        return keycode, False
    # generate the translation map:
    trans = {}
    for keycode, defs in keycodes.items():
        if bool(set(DEBUG_KEYSYMS) & set(bytestostr(d[0]) for d in defs)):
            log.info("client keycode=%i, defs=%s", keycode, defs)
        for bkeysym, i in tuple(defs):             # ie: (b'1', 0) or (b'A', 1), etc
            keysym = bytestostr(bkeysym)
            m = find_keycode(keycode, keysym, i)
            if m:
                x11_keycode, index_matched = m
                trans[(keycode, keysym)] = x11_keycode
                trans[keysym] = x11_keycode
                if index_matched:
                    trans[(keysym, i)] = x11_keycode
    if not xkbmap_x11_keycodes:
        # now add all the keycodes we may not have mapped yet
        # (present in x11_keycodes but not keycodes)
        for keycode, keysyms in x11_keycodes.items():
            for i, keysym in enumerate(keysyms):
                if keysym not in trans:
                    if keysym in DEBUG_KEYSYMS:
                        log.info("x11 keycode %s: %s", keycode, keysym)
                    trans[keysym] = keycode
                key = (keysym, i)
                if key not in trans:
                    if keysym in DEBUG_KEYSYMS:
                        log.info("x11 keycode %s: %s", keycode, key)
                    trans[key] = keycode
    log("set_keycode_translation(..)=%s", Ellipsizer(trans))
    verboselog("set_keycode_translation(..)=%s", trans)
    return trans


def set_all_keycodes(xkbmap_x11_keycodes, xkbmap_keycodes, preserve_server_keycodes, modifiers: dict[str, Iterable]):
    """
        Clients that have access to raw x11 keycodes should provide
        a `xkbmap_x11_keycodes` map, we otherwise fall back to using
        the `xkbmap_keycodes` gtk keycode list.
        We try to preserve the initial keycodes if asked to do so,
        we retrieve them from the current server keymap and combine
        them with the given keycodes.
        The `modifiers` dict can be obtained by calling
        `get_modifiers_from_meanings` or `get_modifiers_from_keycodes`.
        We use it to ensure that two modifiers are not
        mapped to the same keycode (which is not allowed).
        We return a translation map for keycodes after setting them up,
        the key is (keycode, keysym) and the value is the server keycode.
    """
    log("set_all_keycodes(%s.., %s.., %s.., %s)",
        str(xkbmap_x11_keycodes)[:60], str(xkbmap_keycodes)[:60], str(preserve_server_keycodes)[:60], modifiers)

    # so we can validate entries:
    keysym_to_modifier = {}
    for modifier, keysyms in modifiers.items():
        for keysym in keysyms:
            existing_mod = keysym_to_modifier.get(keysym)
            if existing_mod and existing_mod != modifier:
                log.error("ERROR: keysym %s is mapped to both %s and %s !", keysym, modifier, existing_mod)
            else:
                keysym_to_modifier[keysym] = modifier
                if keysym in DEBUG_KEYSYMS:
                    log.info("set_all_keycodes() keysym_to_modifier[%s]=%s", keysym, modifier)
    log("keysym_to_modifier=%s", keysym_to_modifier)

    def modifiers_for(entries) -> set[str]:
        """ entries can only point to a single modifier - verify """
        modifiers: set[str] = set()
        log_fn = log.debug
        for keysym, _ in entries:
            modifier = keysym_to_modifier.get(keysym)
            if modifier:
                modifiers.add(modifier)
            if keysym in DEBUG_KEYSYMS:
                log_fn = log.info
        log_fn("modifiers_for(%s)=%s", entries, modifiers)
        return modifiers

    def filter_mappings(mappings, drop_extra_keys=False) -> dict[int, set]:
        filtered = {}
        invalid_keysyms = set()

        def estr(entries) -> str:
            try:
                return csv(tuple(set(x[0] for x in entries)))
            except Exception:
                return csv(tuple(entries))
        for keycode, entries in mappings.items():
            mods = modifiers_for(entries)
            if len(mods) > 1:
                log.warn("Warning: keymapping changed:")
                log.warn(" keycode %s points to %i modifiers: %s", keycode, len(mods), csv(tuple(mods)))
                log.warn(" from definition: %s", estr(entries))
                for mod in mods:
                    emod = [entry for entry in entries if mod in modifiers_for([entry])]
                    log.warn(" %s: %s", mod, estr(emod))
                # keep just the first one:
                mods = tuple(mods)[:1]
                mod = mods[0]
                entries = [entry for entry in entries if mod in modifiers_for([entry])]
                log.warn(" keeping: %s for %s", estr(entries), mod)
            # now remove entries for keysyms we don't have:
            f_entries = set((keysym, index) for keysym, index in entries
                            if keysym and X11Keyboard.parse_keysym(keysym) is not None)
            for keysym, _ in entries:
                if keysym and X11Keyboard.parse_keysym(keysym) is None:
                    invalid_keysyms.add(keysym)
            if not f_entries:
                log("keymapping removed invalid keycode entry %s pointing to only unknown keysyms: %s",
                    keycode, entries)
                continue
            if drop_extra_keys and not any(keysym for keysym, index in entries if (
                X11Keyboard.parse_keysym(keysym) is not None and keysym not in OPTIONAL_KEYS)
            ):
                log("keymapping removed keycode entry %s pointing to optional keys: %s", keycode, entries)
                continue
            filtered[keycode] = f_entries
        if invalid_keysyms:
            log.warn("Warning: the following keysyms are invalid and have not been mapped")
            log.warn(" %s", csv(invalid_keysyms))
        return filtered

    # get the list of keycodes (either from x11 keycodes or gtk keycodes):
    if xkbmap_x11_keycodes:
        log("using x11 keycodes: %s", xkbmap_x11_keycodes)
        dump_dict(xkbmap_x11_keycodes)
        keycodes = indexed_mappings(xkbmap_x11_keycodes)
    else:
        log("using gtk keycodes: %s", xkbmap_keycodes)
        keycodes = gtk_keycodes_to_mappings(xkbmap_keycodes)
    # filter to ensure only valid entries remain:
    log("keycodes=%s", keycodes)
    keycodes = filter_mappings(keycodes)

    # now lookup the current keycodes (if we need to preserve them)
    preserve_keycode_entries: dict[int, list[str]] = {}
    if preserve_server_keycodes:
        preserve_keycode_entries = X11Keyboard.get_keycode_mappings()
        log("preserved mappings:")
        dump_dict(preserve_keycode_entries)

    kcmin, kcmax = X11Keyboard.get_minmax_keycodes()
    for try_harder in (False, True):
        filtered_preserve_keycode_entries = filter_mappings(indexed_mappings(preserve_keycode_entries), try_harder)
        log("filtered_preserve_keycode_entries=%s", filtered_preserve_keycode_entries)
        trans, new_keycodes, missing_keycodes = translate_keycodes(kcmin, kcmax, keycodes,
                                                                   filtered_preserve_keycode_entries,
                                                                   keysym_to_modifier,
                                                                   try_harder)
        if not missing_keycodes:
            break
    instructions = keymap_to_xmodmap(new_keycodes)
    unset = apply_xmodmap(instructions)
    log("unset=%s", unset)
    return trans


def dump_dict(d) -> None:
    for k, v in d.items():
        log("%s\t\t=\t%s", k, v)


def indexed_mappings(raw_mappings: dict[int, Iterable]) -> dict[int, set]:
    indexed: dict[int, set] = {}
    for keycode, keysyms in raw_mappings.items():
        pairs = set()
        log_fn = log.debug
        for i, keysym in enumerate(keysyms):
            if keysym in DEBUG_KEYSYMS:
                log_fn = log.info
            pairs.add((keysym, i))
        indexed[keycode] = pairs
        log_fn("indexed_mappings: %s=%s", keycode, pairs)
    return indexed


def gtk_keycodes_to_mappings(gtk_mappings: Iterable[tuple[Any, str, int, int, int]]
                             ) -> dict[int, set[tuple[str, int]]]:
    """
        Takes gtk keycodes as obtained by get_gtk_keymap, in the form:
        #[(keyval, keyname, keycode, group, level), ..]
        And returns a list of entries in the form:
        [[keysym, keycode, index], ..]
    """
    # use the keycodes supplied by gtk:
    mappings: dict[int, set[tuple[str, int]]] = {}
    for _, name, keycode, group, level in gtk_mappings:
        if keycode < 0:
            continue            # ignore old 'add_if_missing' client side code
        index = group*2+level
        mappings.setdefault(keycode, set()).add((name, index))
    return mappings


def x11_keycodes_to_list(x11_mappings: dict[int, Iterable]) -> list[tuple[str, int, int]]:
    """
        Takes x11 keycodes as obtained by get_keycode_mappings(), in the form:
        #{keycode : [keysyms], ..}
        And returns a list of entries in the form:
        [[keysym, keycode, index], ..]
    """
    entries = []
    if x11_mappings:
        for keycode, keysyms in x11_mappings.items():
            index = 0
            for keysym in keysyms:
                if keysym:
                    entries.append((keysym, int(keycode), index))
                    if keysym in DEBUG_KEYSYMS:
                        log.info("x11_keycodes_to_list: (%s, %s) : %s", keycode, keysyms, (keysym, int(keycode), index))
                index += 1
    return entries


def translate_keycodes(kcmin: int, kcmax: int, keycodes, preserve_keycode_entries, keysym_to_modifier, try_harder=False):
    """
        The keycodes given may not match the range that the server supports,
        or some of those keycodes may not be usable (only one modifier can
        be mapped to a single keycode) or we want to preserve a keycode,
        or modifiers want to use the same keycode (which is not possible),
        so we return a translation map for those keycodes that have been
        remapped.
        The preserve_keycodes is a dict containing {keycode:[entries]}
        for keys we want to preserve the keycode for.
        Note: a client_keycode of '0' is valid (osx uses that),
        but server_keycode generally starts at 8...
    """
    log("translate_keycodes(%s, %s, %s, %s, %s, %s)",
        kcmin, kcmax, keycodes, preserve_keycode_entries, keysym_to_modifier, try_harder)
    # list of free keycodes we can use:
    free_keycodes = [i for i in range(kcmin, kcmax+1) if i not in preserve_keycode_entries]
    keycode_trans = {}              # translation map from client keycode to our server keycode
    server_keycodes = {}            # the new keycode definitions
    missing_keycodes = []           # the groups of entries we failed to map due to lack of free keycodes

    # to do faster lookups:
    preserve_keysyms_map = {}
    for keycode, entries in preserve_keycode_entries.items():
        for keysym, _ in entries:
            preserve_keysyms_map.setdefault(keysym, set()).add(keycode)
    for k in DEBUG_KEYSYMS:
        log.info("preserve_keysyms_map[%s]=%s", k, preserve_keysyms_map.get(k))

    def do_assign(keycode, server_keycode, entries, override_server_keycode=False) -> int:
        """ may change the keycode if needed
            in which case we update the entries and populate 'keycode_trans'
        """
        log_fn = log.debug
        for name, _ in entries:
            if name in DEBUG_KEYSYMS:
                log_fn = log.info
        if (server_keycode in server_keycodes) and not override_server_keycode:
            log_fn("assign: server keycode %s already in use: %s", server_keycode, server_keycodes.get(server_keycode))
            server_keycode = -1
        elif server_keycode > 0 and (server_keycode < kcmin or server_keycode > kcmax):
            log_fn("assign: keycode %s out of range (%s to %s)", server_keycode, kcmin, kcmax)
            server_keycode = -1
        if server_keycode <= 0:
            if free_keycodes:
                server_keycode = free_keycodes[0]
                log_fn("set_keycodes key %s using free keycode=%s", entries, server_keycode)
            else:
                msg = "set_keycodes: no free keycodes!, cannot translate %s: %s", server_keycode, tuple(entries)
                if try_harder:
                    log.error(*msg)
                else:
                    log_fn(*msg)
                missing_keycodes.append(entries)
                server_keycode = -1
        if server_keycode > 0:
            log_fn("set_keycodes key %s (%s) mapped to keycode=%s", keycode, tuple(entries), server_keycode)
            # can't use it anymore!
            if server_keycode in free_keycodes:
                free_keycodes.remove(server_keycode)
            # record it in trans map:
            for name, _ in entries:
                # noinspection PyChainedComparisons
                if keycode >= 0 and server_keycode != keycode:
                    keycode_trans[(keycode, name)] = server_keycode
                    log_fn(f"keycode_trans[({keycode}, {name})]={server_keycode}")
                keycode_trans[name] = server_keycode
                log_fn(f"keycode_trans[{name}]={server_keycode}")
            server_keycodes[server_keycode] = entries
        return server_keycode

    def assign(client_keycode, entries) -> int:
        if not entries:
            return 0
        # all the keysyms for this keycode:
        keysyms = {keysym for keysym, _ in entries}
        if not keysyms:
            return 0
        if len(keysyms) == 1 and tuple(keysyms)[0] == '0xffffff':
            log("skipped invalid keysym: %s / %s", client_keycode, entries)
            return 0
        log_fn = log.debug
        if [k for k in keysyms if k in DEBUG_KEYSYMS]:
            log_fn = log.info
        log_fn("assign(%s, %s)", client_keycode, entries)

        if not preserve_keycode_entries:
            return do_assign(client_keycode, client_keycode, entries)
        if len(keysyms) == 1:
            # only one keysym, replace with single entry
            # noinspection PySetFunctionToLiteral
            entries = set([(tuple(keysyms)[0], 0)])

        # the candidate preserve entries: those that have at least one of the keysyms:
        preserve_keycode_matches = {}
        for keysym in tuple(keysyms):
            keycodes = preserve_keysyms_map.get(keysym, [])
            for keycode in keycodes:
                v = preserve_keycode_entries.get(keycode)
                preserve_keycode_matches[keycode] = v
                log_fn("preserve_keycode_matches[%s]=%s", keycode, v)

        if not preserve_keycode_matches:
            log_fn("no preserve matches for %s", tuple(entries))
            return do_assign(client_keycode, -1, entries)         # nothing to preserve

        log_fn("preserve matches for %s : %s", entries, preserve_keycode_matches)
        # direct superset:
        for p_keycode, p_entries in preserve_keycode_matches.items():
            if entries.issubset(p_entries):
                log_fn("found direct preserve superset for client keycode %s %s -> server keycode %s %s",
                       client_keycode, tuple(entries), p_keycode, tuple(p_entries))
                return do_assign(client_keycode, p_keycode, p_entries, override_server_keycode=True)
            if p_entries.issubset(entries):
                log_fn("found direct superset of preserve for client keycode %s %s -> server keycode %s %s",
                       client_keycode, tuple(entries), p_keycode, tuple(p_entries))
                return do_assign(client_keycode, p_keycode, entries, override_server_keycode=True)

        # ignoring indexes, but requiring at least as many keysyms:
        for p_keycode, p_entries in preserve_keycode_matches.items():
            p_keysyms = set(keysym for keysym, _ in p_entries)
            if keysyms.issubset(p_keysyms) and len(p_entries) > len(entries):
                log_fn("found keysym preserve superset with more keys for %s : %s",
                       tuple(entries), tuple(p_entries))
                return do_assign(client_keycode, p_keycode, p_entries, override_server_keycode=True)
            if p_keysyms.issubset(keysyms):
                log_fn("found keysym superset of preserve with more keys for %s : %s",
                       tuple(entries), tuple(p_entries))
                return do_assign(client_keycode, p_keycode, entries, override_server_keycode=True)

        if try_harder:
            # try to match the main key only:
            main_key = set((keysym, index) for keysym, index in entries if index == 0)
            if len(main_key) == 1:
                for p_keycode, p_entries in preserve_keycode_matches.items():
                    if main_key.issubset(p_entries):
                        log_fn("found main key superset for %s : %s", main_key, tuple(p_entries))
                        return do_assign(client_keycode, p_keycode, p_entries, override_server_keycode=True)

        log_fn("no matches for %s", tuple(entries))
        return do_assign(client_keycode, -1, entries)

    # now try to assign each keycode:
    for keycode in sorted(keycodes.keys()):
        entries = keycodes.get(keycode)
        log("assign(%s, %s)", keycode, entries)
        assign(keycode, entries)

    # add all the other preserved ones that have not been mapped to any client keycode:
    for server_keycode, entries in preserve_keycode_entries.items():
        if server_keycode not in server_keycodes:
            do_assign(-1, server_keycode, entries)

    # find all keysyms assigned so far:
    all_keysyms = set()
    for entries in server_keycodes.values():
        for x in [keysym for keysym, _ in entries]:
            all_keysyms.add(x)
    log("all_keysyms=%s", all_keysyms)

    # defined keysyms for modifiers if some are missing:
    for keysym, modifier in keysym_to_modifier.items():
        if keysym not in all_keysyms:
            log_fn = log.debug
            if keysym in DEBUG_KEYSYMS:
                log_fn = log.info
            log_fn("found missing keysym %s for modifier %s, will add it", keysym, modifier)
            # noinspection PySetFunctionToLiteral
            new_keycode = set([(keysym, 0)])
            server_keycode = assign(-1, new_keycode)
            log_fn("assigned keycode %s for key '%s' of modifier '%s'", server_keycode, keysym, modifier)

    log("translated keycodes=%s", keycode_trans)
    log("%s free keycodes=%s", len(free_keycodes), free_keycodes)
    return keycode_trans, server_keycodes, missing_keycodes


def keymap_to_xmodmap(trans_keycodes: dict[int, Any]) -> list[tuple]:
    """
        Given a dict with keycodes as keys and lists of keyboard entries as values,
        (keysym, keycode, index)
        produce a list of xmodmap instructions to set the x11 keyboard to match it,
        in the form:
        ("keycode", keycode, [keysyms])
    """
    missing_keysyms = []            # the keysyms lookups which failed
    instructions = []
    all_entries = []
    for entries in trans_keycodes.values():
        all_entries += entries
    keysyms_per_keycode = max([index for _, index in all_entries])+1
    for server_keycode, entries in trans_keycodes.items():
        keysyms = [None]*keysyms_per_keycode
        names = [""]*keysyms_per_keycode
        sentries = sorted(entries, key=lambda x: x[1])
        for name, index in sentries:
            assert 0 <= index < keysyms_per_keycode
            try:
                keysym = X11Keyboard.parse_keysym(name)
            except Exception:
                keysym = None
            if keysym is None:
                if name != "":
                    missing_keysyms.append(name)
            else:
                if keysyms[index] is not None:
                    # if the client provides multiple keysyms for the same index,
                    # replace with the new one if the old one exists elsewhere,
                    # or skip it if we have another entry for it
                    can_override = any(True for i, v in enumerate(keysyms) if i < index and v == keysyms[index])
                    can_skip = any(True for i, v in enumerate(keysyms) if i != index and v == keysym)
                    log_fn = log.debug if can_override or can_skip else log.warn
                    log_fn("Warning: more than one keysym for keycode %-3i at index %i:", server_keycode, index)
                    log_fn(" entries=%s", csv(tuple(sentries)))
                    log_fn(" keysyms=%s", csv(keysyms))
                    log_fn(" assigned keysym=%s", names[index])
                    log_fn(" wanted keysym=%s", name)
                    if can_override:
                        log_fn(" current value also found at another index, overriding it")
                    elif can_skip:
                        log_fn(" new value also found elsewhere, skipping it")
                    else:
                        continue
                names[index] = name
                keysyms[index] = keysym
                if name in DEBUG_KEYSYMS:
                    log.info("keymap_to_xmodmap: keysyms[%s]=%s (%s)", index, keysym, name)
        # remove empty keysyms:
        while keysyms and keysyms[0] is None:
            keysyms = keysyms[1:]
        log_fn = log.debug
        if [k for k in keysyms if k in DEBUG_KEYSYMS]:
            log_fn = log.info
        log_fn("%s: %s -> %s", server_keycode, names, keysyms)
        instructions.append(("keycode", server_keycode, keysyms))

    if missing_keysyms:
        log.error("cannot find the X11 keysym for the following key names: %s", set(missing_keysyms))
    log("instructions=%s", instructions)
    return instructions


################################################################################
# modifiers

def clear_modifiers() -> None:
    instructions = [("clear", i) for i in range(0, 8)]
    apply_xmodmap(instructions)


def set_modifiers(modifiers: dict[str, Iterable[str]]) -> None:
    """
        modifiers is a dict: {modifier : [keynames]}
        Note: the same keysym cannot appear in more than one modifier
    """
    instructions: list[tuple[str, str, Iterable[str]]] = []
    for modifier, keynames in modifiers.items():
        mod = X11Keyboard.parse_modifier(modifier)
        if mod >= 0:
            instructions.append(("add", mod, keynames))
        else:
            log.error("Error: unknown modifier %s", modifier)
    log("set_modifiers: %s", instructions)

    def apply_or_trim(instructions: list[tuple[str, str, Iterable[str]]]) -> None:
        err = apply_xmodmap(instructions)
        log("set_modifiers: err=%s", err)
        if not err:
            return
        log("set_modifiers %s failed, retrying one more at a time", instructions)
        count = len(instructions)
        for i in range(1, count):
            subset = instructions[:i]
            log("set_modifiers testing with [:%s]=%s", i, subset)
            err = apply_xmodmap(subset)
            log("err=%s", err)
            if err:
                log.warn("Warning: removing problematic modifier mapping: %s", csv(instructions[i-1]))
                instructions = instructions[:i-1]+instructions[i:]
                apply_or_trim(instructions)
                return
    apply_or_trim(instructions)


def get_modifiers_from_meanings(xkbmap_mod_meanings: dict[str, str]) -> dict[str, list[str]]:
    """
        xkbmap_mod_meanings maps a keyname to a modifier
        returns keynames_for_mod: {modifier : [keynames]}
    """
    # first generate a {modifier : [keynames]} dict:
    modifiers = {}
    for keyname, modifier in xkbmap_mod_meanings.items():
        keynames = modifiers.setdefault(modifier, [])
        if keyname not in keynames:
            keynames.append(keyname)
    log("get_modifiers_from_meanings(%s) modifier dict=%s", xkbmap_mod_meanings, modifiers)
    return modifiers


def get_modifiers_from_keycodes(xkbmap_keycodes: Iterable, add_default_modifiers=True) -> dict[str, list[str]]:
    """
        Some platforms can't tell us about modifier mappings
        So we try to find matches from the defaults below:
    """
    pref = DEFAULT_MODIFIER_MEANINGS
    # keycodes are: {keycode : (keyval, name, keycode, group, level)}
    matches = {}
    log("get_modifiers_from_keycodes(%s...)", str(xkbmap_keycodes)[:160])
    all_keynames = set()
    for entry in xkbmap_keycodes:
        _, keyname, _, _, _ = entry
        modifier = pref.get(keyname)
        if modifier:
            keynames = matches.setdefault(modifier, [])
            if keyname not in keynames:
                keynames.append(keyname)
            all_keynames.add(keyname)
    if add_default_modifiers:
        # try to add missing ones (magic!)
        defaults = {}
        for keyname, modifier in DEFAULT_MODIFIER_MEANINGS.items():
            if keyname in all_keynames:
                continue            # already defined
            if modifier not in matches:
                # define it since it is completely missing
                keynames = defaults.setdefault(modifier, [])
                if keyname not in keynames:
                    keynames.append(keyname)
            elif modifier in ["shift", "lock", "control", "mod1", "mod2"] or keyname == "ISO_Level3_Shift":
                # these are always added,
                # even if a record for the modifier already exists
                keynames = matches.setdefault(modifier, [])
                if keyname not in keynames:
                    keynames.append(keyname)
        log("get_modifiers_from_keycodes(...) adding defaults: %s", defaults)
        matches.update(defaults)
    log("get_modifiers_from_keycodes(...)=%s", matches)
    return matches


def map_missing_modifiers(keynames_for_mod: dict[str, Iterable]):
    x11_keycodes = X11Keyboard.get_keycode_mappings()
    min_keycode, max_keycode = X11Keyboard.get_minmax_keycodes()
    free_keycodes = [x for x in range(min_keycode, max_keycode) if x not in x11_keycodes]
    log("map_missing_modifiers(%s) min_keycode=%i max_keycode=%i, free_keycodes=%s",
        keynames_for_mod, min_keycode, max_keycode, free_keycodes)
    keysyms_to_keycode = {}
    for keycode, keysyms in x11_keycodes.items():
        for keysym in keysyms:
            keysyms_to_keycode.setdefault(keysym, []).append(keycode)
    xmodmap_changes = []
    for mod, keysyms in keynames_for_mod.items():
        missing = []
        keysym = ""
        for keysym in keysyms:
            if keysym not in keysyms_to_keycode:
                missing.append(keysym)
        if missing:
            log("map_missing_modifiers: no keycode found for modifier keys %s (%s)", csv(missing), mod)
            if not free_keycodes:
                log.warn("Warning: keymap is full, cannot add '%s' for modifier '%s'", keysym, mod)
            else:
                keycode = free_keycodes.pop()
                xmodmap_changes.append(("keycode", keycode, missing))
    if xmodmap_changes:
        log("xmodmap_changes=%s", xmodmap_changes)
        X11Keyboard.set_xmodmap(xmodmap_changes)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys


def get_commonappdata_dir():
    CSIDL_COMMON_APPDATA = 35
    try:
        from ctypes import (
            WinDLL,  # @UnresolvedImport
            create_unicode_buffer,
        )
        from ctypes.wintypes import MAX_PATH
        buf = create_unicode_buffer(MAX_PATH)
        shell32 = WinDLL("shell32", use_last_error=True)
        SHGetFolderPath = shell32.SHGetFolderPathW
        SHGetFolderPath(0, CSIDL_COMMON_APPDATA, None, 0, buf)
        return buf.value
    except Exception:
        return None


def main():
    from multiprocessing import freeze_support
    freeze_support()

    os.environ["XPRA_REDIRECT_OUTPUT"] = "1"
    os.environ["XPRA_LOG_FILENAME"] = "Shadow.log"
    # os.environ["XPRA_ALL_DEBUG"] = "1"
    # os.environ["XPRA_NAMED_PIPE_UNRESTRICTED"] = "1"
    from xpra.platform import init, set_default_name
    set_default_name("Xpra-Shadow")
    init()

    from xpra.scripts.main import main as xpra_main
    args = sys.argv[:1] + [
        "shadow",
    ] + sys.argv[1:]
    commonappdata = get_commonappdata_dir()
    if commonappdata:
        ssl_cert = os.path.join(commonappdata, "Xpra", "ssl-cert.pem")
        if os.path.exists(ssl_cert):
            args.append("--ssl-cert=%s" % ssl_cert)
    sys.exit(xpra_main(sys.argv[0], args))


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# noinspection PyPep8

import os
import sys
import binascii
from enum import Enum, IntEnum
from typing import Final, Protocol, TypeAlias, Any
from collections.abc import Callable, Sized, MutableSequence, Iterable, Sequence

from xpra.util.env import envint, envbool


try:
    # Python 3.11 and later:
    from enum import StrEnum
    from typing import Self
except ImportError:     # pragma: no cover
    StrEnum = Enum      # type: ignore
    Self = Any

try:
    # Python 3.12 and later:
    from collections.abc import Buffer

    class SizedBuffer(Buffer, Sized, Protocol):
        pass
except ImportError:
    class SizedBuffer(Sized, Protocol):
        def __buffer__(self):
            raise NotImplementedError()


PaintCallback: TypeAlias = Callable[[int | bool, str], None]
PaintCallbacks: TypeAlias = MutableSequence[PaintCallback]


ALL_CLIPBOARDS: Final[Sequence[str]] = ("CLIPBOARD", "PRIMARY", "SECONDARY")


# noinspection PyPep8
RESOLUTION_ALIASES: dict[str, tuple[int, int]] = {
    "QVGA"  : (320, 240),
    "VGA"   : (640, 480),
    "SVGA"  : (800, 600),
    "XGA"   : (1024, 768),
    "720P"  : (1280, 720),
    "1080P" : (1920, 1080),
    "FHD"   : (1920, 1080),
    "WQHD"  : (2560, 1440),
    "4K"    : (3840, 2160),
    "5K"    : (5120, 2880),
    "6K"    : (6144, 3456),
    "8K"    : (7680, 4320),
}


def get_default_video_max_size() -> tuple[int, int]:
    svalues = os.environ.get("XPRA_VIDEO_MAX_SIZE", "").replace("x", ",").split(",")
    if len(svalues) == 2:
        try:
            return int(svalues[0]), int(svalues[0])
        except (TypeError, ValueError):
            pass
    return 4096, 4096


def force_size_constraint(width: int, height: int) -> dict[str, dict[str, Any]]:
    size = width, height
    return {
        "size-constraints": {
            "maximum-size": size,
            "minimum-size": size,
            "base-size": size,
        },
    }


VIDEO_MAX_SIZE = get_default_video_max_size()


XPRA_APP_ID: Final[int] = 0
XPRA_GUID1: Final[int] = 0x67b3efa2
XPRA_GUID2: Final[int] = 0xe470
XPRA_GUID3: Final[int] = 0x4a5f
XPRA_GUID4: Final[tuple[int, int, int, int, int, int, int, int]] = (0xb6, 0x53, 0x6f, 0x6f, 0x98, 0xfe, 0x60, 0x81)
XPRA_GUID_STR: Final[str] = "67B3EFA2-E470-4A5F-B653-6F6F98FE6081"
XPRA_GUID_BYTES: Final[bytes] = binascii.unhexlify(XPRA_GUID_STR.replace("-", ""))
XPRA_NOTIFICATIONS_OFFSET: Final[int] = 2**24


class SocketState(StrEnum):
    LIVE = "LIVE"
    DEAD = "DEAD"
    UNKNOWN = "UNKNOWN"
    INACCESSIBLE = "INACCESSIBLE"


# constants shared between client and server:
# (do not modify the values, see also disconnect_is_an_error)
# timeouts:
class ConnectionMessage(StrEnum):
    CLIENT_PING_TIMEOUT     = "client ping timeout"
    LOGIN_TIMEOUT           = "login timeout"
    CLIENT_EXIT_TIMEOUT     = "client exit timeout"
    # errors:
    PROTOCOL_ERROR          = "protocol error"
    VERSION_ERROR           = "version error"
    CONTROL_COMMAND_ERROR   = "control command error"
    AUTHENTICATION_FAILED   = "authentication failed"
    AUTHENTICATION_ERROR    = "authentication error"
    PERMISSION_ERROR        = "permission error"
    SERVER_ERROR            = "server error"
    CONNECTION_ERROR        = "connection error"
    SESSION_NOT_FOUND       = "session not found error"
    # informational (not a problem):
    DONE                    = "done"
    SERVER_EXIT             = "server exit"
    SERVER_UPGRADE          = "server upgrade"
    SERVER_SHUTDOWN         = "server shutdown"
    CLIENT_REQUEST          = "client request"
    DETACH_REQUEST          = "detach request"
    NEW_CLIENT              = "new client"
    IDLE_TIMEOUT            = "idle timeout"
    SESSION_BUSY            = "session busy"
    # client telling the server:
    CLIENT_EXIT             = "client exit"


# convenience method based on the strings above:
def disconnect_is_an_error(reason) -> bool:
    from xpra.util.str_fn import nicestr
    rstr = nicestr(reason)
    return rstr.find("error") >= 0 or (rstr.find("timeout") >= 0 and rstr != ConnectionMessage.IDLE_TIMEOUT.value)


class NotificationID(IntEnum):
    BANDWIDTH   = XPRA_NOTIFICATIONS_OFFSET+1
    IDLE        = XPRA_NOTIFICATIONS_OFFSET+2
    WEBCAM      = XPRA_NOTIFICATIONS_OFFSET+3
    AUDIO       = XPRA_NOTIFICATIONS_OFFSET+4
    OPENGL      = XPRA_NOTIFICATIONS_OFFSET+5
    SCALING     = XPRA_NOTIFICATIONS_OFFSET+6
    NEW_USER    = XPRA_NOTIFICATIONS_OFFSET+7
    CLIPBOARD   = XPRA_NOTIFICATIONS_OFFSET+8
    FAILURE     = XPRA_NOTIFICATIONS_OFFSET+9
    DPI         = XPRA_NOTIFICATIONS_OFFSET+10
    DISCONNECT  = XPRA_NOTIFICATIONS_OFFSET+11
    DISPLAY     = XPRA_NOTIFICATIONS_OFFSET+12
    STARTUP     = XPRA_NOTIFICATIONS_OFFSET+13
    FILETRANSFER    = XPRA_NOTIFICATIONS_OFFSET+14
    SHADOWWAYLAND   = XPRA_NOTIFICATIONS_OFFSET+15


class Gravity(IntEnum):
    # X11 constants we use for gravity:
    NorthWest   = 1
    North       = 2
    NorthEast   = 3
    West        = 4
    Center      = 5
    East        = 6
    SouthWest   = 7
    South       = 8
    SouthEast   = 9
    Static      = 10


def gravity_str(v) -> str:
    try:
        return Gravity(v).name
    except ValueError:
        return str(v)


# initiate-moveresize X11 constants
class MoveResize(IntEnum):
    SIZE_TOPLEFT      = 0
    SIZE_TOP          = 1
    SIZE_TOPRIGHT     = 2
    SIZE_RIGHT        = 3
    SIZE_BOTTOMRIGHT  = 4
    SIZE_BOTTOM       = 5
    SIZE_BOTTOMLEFT   = 6
    SIZE_LEFT         = 7
    MOVE              = 8
    SIZE_KEYBOARD     = 9
    MOVE_KEYBOARD     = 10
    CANCEL            = 11


MOVERESIZE_DIRECTION_STRING = {
    MoveResize.SIZE_TOPLEFT      : "SIZE_TOPLEFT",
    MoveResize.SIZE_TOP          : "SIZE_TOP",
    MoveResize.SIZE_TOPRIGHT     : "SIZE_TOPRIGHT",
    MoveResize.SIZE_RIGHT        : "SIZE_RIGHT",
    MoveResize.SIZE_BOTTOMRIGHT  : "SIZE_BOTTOMRIGHT",
    MoveResize.SIZE_BOTTOM       : "SIZE_BOTTOM",
    MoveResize.SIZE_BOTTOMLEFT   : "SIZE_BOTTOMLEFT",
    MoveResize.SIZE_LEFT         : "SIZE_LEFT",
    MoveResize.MOVE              : "MOVE",
    MoveResize.SIZE_KEYBOARD     : "SIZE_KEYBOARD",
    MoveResize.MOVE_KEYBOARD     : "MOVE_KEYBOARD",
    MoveResize.CANCEL            : "CANCEL",
}
SOURCE_INDICATION_UNSET: Final[int] = 0
SOURCE_INDICATION_NORMAL: Final[int] = 1
SOURCE_INDICATION_PAGER: Final[int] = 2
SOURCE_INDICATION_STRING: Final[dict[int, str]] = {
    SOURCE_INDICATION_UNSET      : "UNSET",
    SOURCE_INDICATION_NORMAL     : "NORMAL",
    SOURCE_INDICATION_PAGER      : "PAGER",
}


# magic value for "workspace" window property, means unset
WORKSPACE_UNSET: Final[int] = 65535
WORKSPACE_ALL: Final[int] = 0xffffffff
WORKSPACE_NAMES = {
    WORKSPACE_UNSET  : "unset",
    WORKSPACE_ALL    : "all",
}


CLOBBER_UPGRADE: int = 0x1
CLOBBER_USE_DISPLAY: int = 0x2

# if you want to use a virtual screen bigger than this
# you will need to change those values, but some broken toolkits
# will then misbehave (they use signed shorts instead of signed ints..)
MAX_WINDOW_SIZE: int = 2**15-2**13


GROUP: str = os.environ.get("XPRA_GROUP", "xpra")

FULL_INFO: int = envint("XPRA_FULL_INFO", 1)
assert FULL_INFO >= 0
LOG_HELLO: bool = envbool("XPRA_LOG_HELLO", False)

SSH_AGENT_DISPATCH: bool = envbool("XPRA_SSH_AGENT_DISPATCH", os.name == "posix")

MIN_COMPRESS_SIZE: int = envint("XPRA_MIN_DECOMPRESSED_SIZE", -1)
MAX_DECOMPRESSED_SIZE: int = envint("XPRA_MAX_DECOMPRESSED_SIZE", 256*1024*1024)


MIN_DPI: int = envint("XPRA_MIN_DPI", 10)
MAX_DPI: int = envint("XPRA_MIN_DPI", 500)

SYNC_ICC: bool = envbool("XPRA_SYNC_ICC", True)

DEFAULT_REFRESH_RATE: int = envint("XPRA_DEFAULT_REFRESH_RATE", 50*1000)

SPLASH_EXIT_DELAY: int = envint("XPRA_SPLASH_EXIT_DELAY", 4)

DEFAULT_XDG_DATA_DIRS: str = ":".join(
    (
        "/usr/share",
        "/usr/local/share",
        "~/.local/share/applications",
        "~/.local/share/flatpak/exports/share",
        "/var/lib/flatpak/exports/share",
    )
)


def noop(*_args, **_kwargs) -> None:
    """ do nothing """


WINDOW_DECODE_SKIPPED: Final[int] = 0
WINDOW_DECODE_ERROR: Final[int] = -1
WINDOW_NOT_FOUND: Final[int] = -2


ScreenshotData = tuple[int, int, str, int, bytes]


MIN_VREFRESH = envint("XPRA_MIN_VREFRESH", 5)
MAX_VREFRESH = envint("XPRA_MAX_VREFRESH", 144)


def i(value, default: int) -> int:
    try:
        return int(value)
    except ValueError:
        return default


def get_refresh_rate_for_value(refresh_rate: str, invalue: int, multiplier=1) -> int:
    if refresh_rate.lower() in ("none", "auto"):
        # just leave it unchanged:
        return invalue
    # refresh rate can be a value (ie: 40), or a range (10-50)
    parts = refresh_rate.split("-", 1)
    if len(parts) > 1:
        minvr = i(parts[0], MIN_VREFRESH) * multiplier
        maxvr = i(parts[1], MAX_VREFRESH) * multiplier
        if minvr > maxvr:
            raise ValueError("the minimum refresh rate cannot be greater than the maximum")
        return min(maxvr, max(minvr, invalue))

    minvr = MIN_VREFRESH * multiplier
    maxvr = MAX_VREFRESH * multiplier
    if refresh_rate.endswith("%") > 0:
        mult = int(refresh_rate[:-1])  # ie: "80%" -> 80
        value = round(invalue * mult / 100)
        return min(maxvr, max(minvr, value))

    try:
        value = int(refresh_rate) * multiplier
        return min(maxvr, max(minvr, value))
    except ValueError:
        pass
    # fallback to client supplied value:
    return invalue


def adjust_monitor_refresh_rate(refresh_rate: str, mdef) -> dict[int, dict]:
    adjusted = {}
    for i, monitor in mdef.items():
        # make a copy, don't modify in place!
        # (as this may be called multiple times on the same input dict)
        mprops = dict(monitor)
        if refresh_rate != "auto":
            value = int(monitor.get("refresh-rate", DEFAULT_REFRESH_RATE))
            value = get_refresh_rate_for_value(refresh_rate, value, 1000)
            if value:
                mprops["refresh-rate"] = value
        adjusted[i] = mprops
    return adjusted


# this default value is based on 0.15.x clients,
# later clients should provide the `metadata.supported` capability instead
DEFAULT_METADATA_SUPPORTED = (
    "title", "icon-title", "pid", "iconic",
    "size-hints", "class-instance", "client-machine",
    "transient-for", "window-type",
    "fullscreen", "maximized", "decorations", "skip-taskbar", "skip-pager",
    "has-alpha", "override-redirect", "tray", "modal",
    "role", "opacity", "xid", "group-leader",
    "opaque-region",
    "command", "workspace", "above", "below", "sticky",
    "set-initial-position", "requested-position",
    "content-type",
    # 4.4:
    # "parent", "relative-position",
)


def noerr(fn: Callable, *args):
    # noinspection PyBroadException
    try:
        return fn(*args)
    except Exception:
        return None


def roundup(n: int, m: int) -> int:
    return (n + m - 1) & ~(m - 1)


def uniq(seq: Iterable) -> list:
    seen = set()
    return [x for x in seq if not (x in seen or seen.add(x))]


def get_run_info(subcommand="server") -> Sequence[str]:
    from xpra.os_util import POSIX
    from xpra.util.version import full_version_str, get_platform_info
    from xpra.util.system import platform_name
    from xpra.log import Logger
    log = Logger("util")
    run_info = [f"xpra {subcommand} version {full_version_str()}"]
    try:
        pinfo = get_platform_info()
        osinfo = " on " + platform_name(sys.platform,
                                        pinfo.get("linux_distribution") or pinfo.get("sysrelease", ""))
    except OSError:
        log("platform name error:", exc_info=True)
        osinfo = ""
    if POSIX:
        uid = os.getuid()
        gid = os.getgid()
        try:
            import pwd
            import grp

            user = pwd.getpwuid(uid)[0]
            group = grp.getgrgid(gid)[0]
            run_info.append(f" {uid=} ({user}), {gid=} ({group})")
        except (TypeError, KeyError):
            log("failed to get user and group information", exc_info=True)
            run_info.append(f" {uid=}, {gid=}")
    run_info.append(" running with pid %s%s" % (os.getpid(), osinfo))
    vinfo = ".".join(str(x) for x in sys.version_info[:FULL_INFO + 1])
    run_info.append(f" {sys.implementation.name} {vinfo}")
    return run_info
